# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional, Tuple, Union

import numpy as np

import nncf
from nncf.common.logging.logger import nncf_logger
from nncf.common.utils.backend import is_openvino_available
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.fake_quantize import calculate_scale_zero_point
from nncf.tensor import Tensor
from nncf.tensor import functions as fns
from nncf.tensor.definitions import TensorBackend
from nncf.tensor.definitions import TensorDataType

ReductionAxes = Union[int, Tuple[int, ...]]

NF4_QUANTILES = np.array(
    [
        -1.0,
        -0.6961928009986877,
        -0.5250730514526367,
        -0.39491748809814453,
        -0.28444138169288635,
        -0.18477343022823334,
        -0.09105003625154495,
        0.0,
        0.07958029955625534,
        0.16093020141124725,
        0.24611230194568634,
        0.33791524171829224,
        0.44070982933044434,
        0.5626170039176941,
        0.7229568362236023,
        1.0,
    ],
    dtype=np.float32,
)

CENTER_OF_NF4_QUANTILES = (NF4_QUANTILES[1:] + NF4_QUANTILES[:-1]) / 2


@dataclass
class CompressedWeight:
    """
    Compressed weight and decompression parameters.

    :param tensor: The tensor with compressed weight.
    :param scale: The decompression scale, in practice it is dequantization scale for the INT quantization.
    :param zero_point: The zero-point, it is the value of the compression type corresponding to the value 0
        in the non-compression realm. Applicable for INT quantization.
    """

    tensor: Tensor
    scale: Tensor
    zero_point: Optional[Tensor] = None


def reshape_weight_for_grouped_quantization(
    weight: Tensor, reduction_axes: ReductionAxes, group_size: int
) -> Tuple[Tensor, int]:
    """
    Reshapes weight for group-wise quantization and return a reduction axis for collecting statistics per group
    dimension. Having a transposed weight with shapes [c_out, c_in] and group size = 128, shape of reshaped weight is
    [c_out, c_in // 128, 128], reduction axis = 1 and the returned reduction axis = 2.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param group_size: Number of weights (e.g. 128) in the channel dimension that share quantization parameters (scale).
    :return: reshaped weight and new reduction axis.
    """
    assert group_size != -1
    if isinstance(reduction_axes, tuple) and len(reduction_axes) == 1:
        reduction_axes = reduction_axes[0]
    if not isinstance(reduction_axes, int):
        raise nncf.UnsupportedModelError(
            f"Group-wise quantization expects a single reduction axis, but given: {reduction_axes}."
        )
    channel_size = weight.shape[reduction_axes]
    if channel_size % group_size != 0:
        raise nncf.UnsupportedModelError(
            f"Channel size {channel_size} should be divisible by size of group {group_size}"
        )

    num_groups_per_channel = channel_size // group_size
    shape = list(weight.shape)  # [a1, r, a2] - "r" refers to number of channels along reduction axis
    shape[reduction_axes : reduction_axes + 1] = (num_groups_per_channel, group_size)
    reshaped_weight = weight.reshape(shape)
    reduction_axes += 1
    return reshaped_weight, reduction_axes


def calculate_nf4_scale(weight: Tensor, reduction_axes: ReductionAxes) -> Tensor:
    """
    Calculates the scale for nf4 quantization.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes along which to reduce (collect) different statistics (e.g., min, max).
    :return: Scale tensor of float32 type for nf4 quantization.
    """
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    scale = fns.max(fns.abs(weight), axis=reduction_axes, keepdims=True)

    # NOTE: adding machine epsilon to avoid division by zero
    eps = fns.finfo(weight).eps
    scale = fns.where(fns.abs(scale) < eps, eps, scale)

    return scale


def calculate_e2m1_scale(weight: Tensor, reduction_axes: ReductionAxes, max_val=6.0) -> Tensor:
    """
    Calculates the scale for e2m1 quantization.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes along which to reduce (collect) different statistics (e.g., min, max).
    :param max_val: Maximal value of e2m1 type.
    :param to_e8m0: Defines convert scale to e8m0 or not.
    :return: Scale tensor of float32 type for e2m1 quantization.
    """
    scale = calculate_nf4_scale(weight, reduction_axes) / max_val

    scale = fns.log2(scale)
    scale = fns.ceil(scale)
    scale = fns.clip(scale, -127, 127)
    scale = 2**scale

    return scale


def calculate_signed_scale(weight: Tensor, reduction_axes: ReductionAxes, num_bits=4) -> Tensor:
    """
    Calculates the signed scale for symmetric quantization.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes along which to reduce (collect) different statistics (e.g., min, max).
    :param num_bits: number of bits in compression.
    :return: Scale tensor.
    """
    level_high = 2 ** (num_bits - 1)

    w_abs_min = fns.abs(fns.min(weight, axis=reduction_axes, keepdims=True))
    w_max = fns.max(weight, axis=reduction_axes, keepdims=True)

    scale = fns.where(w_abs_min >= w_max, w_abs_min, -w_max)
    scale /= level_high

    eps = fns.finfo(scale).eps
    scale = fns.where(fns.abs(scale) < eps, eps, scale)

    return scale


def calculate_normalized_weight(weight: Tensor, scale: Tensor) -> Tensor:
    """
    Normalizes the weight tensor using the provided scale.

    :param weight: Weight tensor to normalize.
    :param scale: Scale tensor used for normalization.
    :return: Normalized weight tensor.
    """
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)
    if scale.dtype != TensorDataType.float32:
        scale = scale.astype(TensorDataType.float32)

    return weight / scale


def do_nf4_quantization(weight: Tensor, scale: Tensor, is_normalized_weight: bool = False) -> Tensor:
    """
    Performs NF4 quantization. The floating point values are represented by floating point scale and look-up with
    16 floating-point values on [-1, 1]. Scale normalizes original values to [-1, 1] interval and look-up table
    "rounds" or "quantize" to the closest quant.

    :param weight: Weight tensor to quantize.
    :param scale: Scale tensor used for normalization.
    :param is_normalized_weight: Whether weight was scaled to [-1, 1] interval. Defaults to False.
    :return: Tensor with floating-point values, where each of them corresponds to 1 out of 16 quants on [-1, 1].
    """
    norm_weight = weight if is_normalized_weight else calculate_normalized_weight(weight, scale)
    center_nf4_quantiles = fns.from_numpy(CENTER_OF_NF4_QUANTILES, backend=norm_weight.backend)
    indexes = fns.searchsorted(center_nf4_quantiles, norm_weight)
    nf4_quantiles = fns.from_numpy(NF4_QUANTILES, backend=indexes.backend)
    nf4_weight = nf4_quantiles[indexes]
    return nf4_weight


def do_nf4_dequantization(nf4_weight: Tensor, scale: Tensor, reduction_axis: int = -1) -> Tensor:
    """
    Decompresses the NF4 quantized weight tensor.

    :param nf4_weight: Tensor with floating-point values,
        where each of them corresponds to 1 out of 16 quants on [-1, 1].
    :param scale: Scale tensor used for decompression.
    :param reduction_axis: axis along which weights were reshaped for group quantization and will be reshaped back to
        original shapes. If equals to -1, weights are not reshaped, assumed not a group quantization. Defaults to -1.
    :return: Decompressed weight tensor.
    """
    decompressed_weight = nf4_weight * scale
    if reduction_axis != -1:
        decompressed_weight = ungroup_weights(decompressed_weight, reduction_axis)
    return decompressed_weight


def calculate_normalized_weight_and_fp4_scale(
    weight: Tensor,
    reduction_axes: ReductionAxes,
    group_size: int = -1,
    precomputed_scale: Tensor = None,
    mode: CompressWeightsMode = CompressWeightsMode.NF4,
) -> Tuple[Tensor, Tensor]:
    """
    Calculates scale for fp4 (nf4, e2m1) quantization and normalizes weights by the scale.
    Weights are reshaped in case of positive value of group size.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param group_size: Number of weights (e.g. 128) in the channel dimension that share quantization parameters (scale).
        The value -1 means no grouping. Defaults to -1.
    :param precomputed_scale: Precomputed scale.
    :return: Normalized weight tensor of float32 type and nf4 scale tensor of float32 type.
    """
    assert mode in [CompressWeightsMode.NF4, CompressWeightsMode.E2M1]
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    if group_size != -1:
        # weights are reshaped: [a1, r, a2] -> [a1, r//gs, gs, a2]
        weight, reduction_axes = reshape_weight_for_grouped_quantization(weight, reduction_axes, group_size)

    if mode == CompressWeightsMode.NF4:
        scale = calculate_nf4_scale(weight, reduction_axes) if precomputed_scale is None else precomputed_scale
    if mode == CompressWeightsMode.E2M1:
        scale = calculate_e2m1_scale(weight, reduction_axes) if precomputed_scale is None else precomputed_scale
    norm_weight = calculate_normalized_weight(weight, scale)
    return norm_weight, scale


def calculate_integer_quantization_params(
    weight: Tensor,
    reduction_axes: ReductionAxes,
    config: WeightCompressionConfig,
) -> Tuple[Tensor, Tensor]:
    """
    Calculates the scale and zero point for uniform quantization (INT4, INT8), when the range of values is divided into
    equal intervals, and each interval is assigned a quant.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param config: Weight compression configuration.
    :return: Scale and zero point tensors.
    """
    if not config.is_integer:
        raise nncf.InternalError("The function supports integer quantization only")

    num_bits = config.num_bits

    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    if config.is_asym_mode:
        level_low = 0
        level_high = 2**num_bits - 1
        min_values = fns.min(weight, axis=reduction_axes, keepdims=True)  # [a1, r, a2] -> [a1, 1, a2]
        max_values = fns.max(weight, axis=reduction_axes, keepdims=True)  # [a1, r, a2] -> [a1, 1, a2]
        scale, zero_point = calculate_scale_zero_point(
            min_values, max_values, level_low, level_high, narrow_range=False
        )
        return scale, zero_point

    scale = calculate_signed_scale(weight, reduction_axes, num_bits)
    return scale, None


def calculate_quantized_weight(
    weight: Tensor,
    config: WeightCompressionConfig,
    scale: Tensor,
    zero_point: Optional[Tensor] = None,
) -> Tensor:
    """
    Quantizes the weight tensor using the provided scale and zero point.

    :param weight: Weight tensor to quantize.
    :param config: Weight compression configuration.
    :param scale: Scale tensor used for quantization.
    :param zero_point: Zero point tensor used for quantization.
    :return: Quantized weight tensor of uint8 or int8 type.
    """
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)
    if scale.dtype != TensorDataType.float32:
        scale = scale.astype(TensorDataType.float32)

    num_bits = config.num_bits
    asym_quant = config.is_asym_mode
    dtype = TensorDataType.uint8 if asym_quant else TensorDataType.int8
    level_low = 0 if asym_quant else -(2 ** (num_bits - 1))
    level_high = 2**num_bits - 1 if asym_quant else 2 ** (num_bits - 1) - 1

    compressed_weights = weight / scale
    if zero_point is not None:
        compressed_weights += zero_point.astype(weight.dtype)
    compressed_weights = fns.round(compressed_weights)
    compressed_weights = fns.clip(compressed_weights, level_low, level_high).astype(dtype)

    return compressed_weights


def get_integer_quantization_error(
    weight: Tensor,
    reduction_axes: ReductionAxes,
    config: WeightCompressionConfig,
) -> float:
    """
    Calculates a quantity characterizing the difference between floating point weights and fake quantized
    (compressed and decompressed) to integer ones.

    :param weight: Weight array to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param config: Information on how to compress (quantize) a specific weight.
    :return: The quantity characterizing the error of integer quantization.
    """
    orig_shape = weight.shape

    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    decompressed_weight = quantize_dequantize_weight(weight, config, reduction_axes)

    decompressed_weight = decompressed_weight.reshape(orig_shape)
    diff = (decompressed_weight - weight) ** 2
    layer_err = fns.mean(diff, axis=reduction_axes)
    val = fns.max(layer_err)
    return val.item()


def compress_weight(
    weight: Tensor,
    reduction_axes: ReductionAxes,
    config: WeightCompressionConfig,
    precomputed_scale: Tensor = None,
    precomputed_zero_point: Tensor = None,
):
    """
    Compress weight using compression configuration.

    :param weight: The weight to compress.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param config: Compression configuration.
    :param precomputed_scale: Precomputed scale.
    :param precomputed_zero_point: Precomputed zero point.
    :return: The compressed weight and decompression parameters as instance of CompressedWeight
    """
    if not config.is_integer:
        if weight.backend == TensorBackend.ov:
            weight = weight.as_numpy_tensor()

        compressed_weight, scale = calculate_normalized_weight_and_fp4_scale(
            weight, reduction_axes, config.group_size, precomputed_scale, config.mode
        )
        return CompressedWeight(compressed_weight, scale)
    compressed_weight, scale, zero_point = do_int_quantization(
        weight, config, reduction_axes, precomputed_scale, precomputed_zero_point
    )

    return CompressedWeight(compressed_weight, scale, zero_point)


def ungroup_weights(weights: Tensor, reduction_axis: int) -> Tensor:
    """
    Reshapes weights used for group quantization back to original shape.

    :param weights: The weight to reshape.
    :param reduction_axis: The axis, along which weights were reshaped for group quantization and will be reshaped back
        to original shapes. If equals to -1, weights are not reshaped, assumed not a group quantization. Default to -1.
    :return: Reshaped weight.
    """
    shape = list(weights.shape)  # [a1, r, a2] - "r" refers to number of channels along reduction axis
    shape[reduction_axis] = shape[reduction_axis] * shape[reduction_axis + 1]
    shape[reduction_axis + 1] = 1
    reshaped_weight = weights.reshape(shape)
    reshaped_weight = fns.squeeze(reshaped_weight)
    weights = reshaped_weight
    return weights


def do_int_dequantization(
    compressed_weights: Tensor, scale: Tensor, zero_point: Optional[Tensor] = None, reduction_axis: int = -1
) -> Tensor:
    """
    The method dequantizes the given weights to float point data type in accordance with the scale and
    zero_point data type.

    :param compressed_weights: compressed weights.
    :param scale: scale in compression/quantization.
    :param zero_point: zero point in compression/quantization.
    :param reduction_axis: axis along which weights were reshaped for group quantization and will be reshaped back to
        original shapes. If equals to -1: weights are not reshaped, assumed not a group quantization. Default to -1.
    :return: dequantized/decompressed weights.
    """
    decompressed_weight = (
        compressed_weights.astype(TensorDataType.int32) - zero_point if zero_point is not None else compressed_weights
    )
    decompressed_weight = decompressed_weight.astype(scale.dtype) * scale

    if reduction_axis > -1:
        decompressed_weight = ungroup_weights(decompressed_weight, reduction_axis)

    return decompressed_weight


def do_int_quantization(
    weight: Tensor,
    config: WeightCompressionConfig,
    reduction_axes: Optional[ReductionAxes] = None,
    precomputed_scale: Tensor = None,
    precomputed_zero_point: Tensor = None,
    **kwargs,
) -> Tuple[Tensor, Tensor, Tensor]:
    """
    Performs integer quantization on the given weight tensor.

    :param weight: The weight tensor to quantize.
    :param config: The weight compression configuration.
    :param reduction_axes: Axes along which to reduce (collect) statistics (e.g., min, max). Not required if
        precomputed scale (and zero point) are provided.
    :param precomputed_scale: Optional precomputed scale tensor.
    :param precomputed_zero_point: Optional precomputed zero point tensor.
    :return: A tuple containing the compressed weights, scale, and zero point tensors.
    """
    if not config.is_integer:
        raise nncf.InternalError("The function supports integer quantization only")
    if config.is_asym_mode and (precomputed_scale is None) != (precomputed_zero_point is None):
        raise ValueError(
            "If precomputed quantization parameters are provided, both scale and zero point are required "
            "for asymmetric quantization."
        )

    # When reduction axes are not provided, assuming that the weights are already reshaped
    if config.group_size != -1 and reduction_axes is not None:
        # weights are reshaped from [a1, r, a2] to [a1, r//gs, gs, a2]
        weight, reduction_axes = reshape_weight_for_grouped_quantization(weight, reduction_axes, config.group_size)

    # Optimized implementation
    if is_openvino_available() and weight.backend in [TensorBackend.ov, TensorBackend.numpy]:
        from nncf.openvino.optimized_functions import do_int_quantization as do_int_quantization_ov

        return do_int_quantization_ov(
            weight, config, reduction_axes, precomputed_scale, precomputed_zero_point, **kwargs
        )
    if not is_openvino_available() and weight.backend in [TensorBackend.ov, TensorBackend.numpy]:
        nncf_logger.info_once(
            "OpenVINO optimizations are disabled. Install OpenVINO to enable them and improve the performance."
        )

    # Reference implementation
    if weight.backend == TensorBackend.ov:
        weight = weight.as_numpy_tensor()
    if weight.dtype != TensorDataType.float32:
        weight = weight.astype(TensorDataType.float32)

    scale, zero_point = None, None
    if precomputed_scale is None or (config.is_asym_mode and precomputed_zero_point is None):
        if reduction_axes is None:
            raise ValueError("Reduction axes are required for computing the scale and (zero point) vectors.")
        scale, zero_point = calculate_integer_quantization_params(weight, reduction_axes, config)
    if precomputed_scale is not None:
        scale = precomputed_scale
    if precomputed_zero_point is not None:
        zero_point = precomputed_zero_point

    compressed_weights = calculate_quantized_weight(weight, config, scale, zero_point)
    return compressed_weights, scale, zero_point


def quantize_dequantize_weight(
    weight: Tensor,
    config: WeightCompressionConfig,
    reduction_axes: Optional[ReductionAxes] = None,
    precomputed_scale: Optional[Tensor] = None,
    precomputed_zero_point: Optional[Tensor] = None,
    return_compressed_weight: Optional[bool] = False,
    **kwargs,
) -> Union[Tensor, Tuple[Tensor, Tensor, Tensor, Tensor]]:
    """
    First quantizes the given weight tensor and then dequantizes it back to obtain float32 values.
    :param weight: The weight tensor to quantize-dequantize.
    :param config: Compression configuration.
    :param reduction_axes: Axes along which to reduce (collect) statistics (e.g., min, max). Not required if
        precomputed scale (and zero point) are provided.
    :param precomputed_scale: Optional precomputed scale tensor.
    :param precomputed_zero_point: Optional precomputed zero point tensor.
    :param return_compressed_weight: If True, besides decompressed weight will also return compressed weight, scale,
        (and zero point).
    :return: Dequantized weight tensor or a tuple containing the decompressed weight, compressed weight, scale,
        (and zero point).
    """
    # Optimized implementation
    if is_openvino_available() and weight.backend in [TensorBackend.ov, TensorBackend.numpy]:
        from nncf.openvino.optimized_functions import quantize_dequantize_weight as quantize_dequantize_weight_ov

        return quantize_dequantize_weight_ov(
            weight,
            config,
            reduction_axes,
            precomputed_scale,
            precomputed_zero_point,
            return_compressed_weight,
            **kwargs,
        )
    if not is_openvino_available() and weight.backend in [TensorBackend.ov, TensorBackend.numpy]:
        nncf_logger.info_once(
            "OpenVINO optimizations are disabled. Install OpenVINO to enable them and improve the performance."
        )

    # Reference implementation
    compressed_weight, scale, zero_point = do_int_quantization(
        weight, config, reduction_axes, precomputed_scale, precomputed_zero_point
    )
    decompressed_weight = do_int_dequantization(compressed_weight, scale, zero_point)
    if return_compressed_weight:
        return decompressed_weight, compressed_weight, scale, zero_point
    else:
        return decompressed_weight



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.definitions import CONST_SPARSITY_ALGO_NAME_IN_CONFIG
from nncf.config.schemata.common.compression import BASIC_COMPRESSION_ALGO_SCHEMA
from nncf.config.schemata.common.targeting import SCOPING_PROPERTIES

CONST_SPARSITY_SCHEMA = {
    **BASIC_COMPRESSION_ALGO_SCHEMA,
    "properties": {
        "algorithm": {"const": CONST_SPARSITY_ALGO_NAME_IN_CONFIG},
        **SCOPING_PROPERTIES,
    },
    "additionalProperties": False,
    "description": "This algorithm takes no additional parameters and is used when you want to load "
    "a checkpoint trained with another sparsity algorithm and do other compression without "
    "changing the sparsity mask.",
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractclassmethod
from abc import abstractmethod
from typing import Any, Dict

import torch
from torch import nn

from nncf.common.hook_handle import HookHandle
from nncf.common.hook_handle import add_op_to_registry
from nncf.common.utils.registry import Registry

COMPRESSION_MODULES = Registry("compression modules")


class StatefullModuleInterface(ABC):
    """
    Interface that should be implemented for every registered compression module to make it possible
    to save an compression modules state and create an compression module from the saved state.
    Config of the module should be json serializable, no python objects except
    standart (str, list and etc.) should be present in a compression module config.
    Values for attributes with type torch.nn.Parameter
    is recovered from the model `state_dict`, so there is no need to keep them in the module config.
    Modules should avoid implementation of `__call__` method and use `forward` method instead,
    as torch functions called inside the `__call__` method could not be unambiguously
    separated from the wrapped parent nncf module functions calls, thus nncf is unable to
    identify target point for that call during transformations recovery process.
    """

    @abstractmethod
    def get_config(self) -> Dict[str, Any]:
        """
        Returns the compression module config.
        """

    @abstractclassmethod
    def from_config(cls, state: Dict[str, Any]) -> object:
        """
        Creates a compression module instance from the given config.
        """


class ProxyModule:
    def __init__(self, module):
        self._module = module

    def __getattr__(self, name):
        return getattr(self._module, name)

    @property
    def __class__(self):
        return type(self._module)


class _NNCFModuleMixin:
    """
    Default class for modules that will be optimized by NNCF.

        Attributes:
            op_func_name    Name of corresponding torch function.
            target_weight_dim_for_compression   Target dimension of weights that will be compressed in some algorithms.
            ignored_algorithms   List of algorithms that will skip the module.
            _custom_forward_fn  wrapper of the custom forward function that is called with `self` argument equals to the
                ProxyModule
    """

    op_func_name = ""
    target_weight_dim_for_compression = 0
    _custom_forward_fn = None
    ignored_algorithms = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _NNCFModuleMixin.add_mixin_fields(self)

    @staticmethod
    def add_mixin_fields(obj):
        obj.pre_ops = nn.ModuleDict()
        obj.post_ops = nn.ModuleDict()

    def get_pre_op(self, key):
        return self.pre_ops[key]

    def get_post_op(self, key):
        return self.post_ops[key]

    def register_pre_forward_operation(self, op) -> HookHandle:
        return add_op_to_registry(self.pre_ops, op)

    def remove_pre_forward_operation(self, key):
        return self.pre_ops.pop(key)

    def register_post_forward_operation(self, op) -> HookHandle:
        return add_op_to_registry(self.post_ops, op)

    def remove_post_forward_operation(self, key):
        return self.post_ops.pop(key)

    def reset(self):
        self.pre_ops.clear()
        self.post_ops.clear()

    def forward(self, *args):
        proxy_module = ProxyModule(self)
        for op in self.pre_ops.values():
            op_args = op(proxy_module, args)
            if op_args is not None:
                if not isinstance(op_args, tuple):
                    op_args = tuple([op_args])
                args = op_args
        forward_fn = self._custom_forward_fn.__func__ if self._custom_forward_fn else super().forward.__func__
        results = forward_fn(proxy_module, *args)
        for op in self.post_ops.values():
            op_results = op(proxy_module, results)
            if op_results is not None:
                results = op_results
        return results


class CompressionParameter(nn.Parameter):
    """
    The class that should be used in all compression algorithms instead of torch.nn.Parameter.

    This class utilize `compression_lr_multiplier` parameter from :class:`nncf.NNCFConfig`
    to increase/decrease gradients for compression algorithms' parameters.
    """

    def __new__(cls, data: torch.Tensor = None, requires_grad: bool = True, compression_lr_multiplier: float = None):
        return super().__new__(cls, data, requires_grad=requires_grad)

    def __init__(self, data: torch.Tensor = None, requires_grad: bool = True, compression_lr_multiplier: float = None):
        """

        Args:
            data: Parameter tensor
            requires_grad: If the parameter requires gradient
            compression_lr_multiplier: Multiplier for gradient values
        """
        super().__init__()

        if compression_lr_multiplier is not None and self.dtype.is_floating_point:
            self.requires_grad = True
            self.register_hook(lambda grad: compression_lr_multiplier * grad)
            self.requires_grad = requires_grad



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.experimental.common.graph.netron import save_for_netron
from nncf.experimental.common.pruning.block_hierarchy import BlockHierarchy
from nncf.experimental.common.pruning.propagation_data import ConsumerInfo
from nncf.experimental.common.pruning.propagation_data import ProducerInfo
from nncf.experimental.common.pruning.propagation_data import PropagationGroup
from nncf.experimental.common.pruning.propagation_data import PropagationMask
from nncf.experimental.common.pruning.propagation_data import PruningBlock


@dataclass
class PruningGroup:
    """
    Group of pruning blocks that is obtained after propagation.
    """

    block: PruningBlock
    producers: Set[ProducerInfo]
    consumers: Set[ConsumerInfo]

    def __eq__(self, other: "PruningGroup"):
        return self.block == other.block and self.producers == other.producers and self.consumers == other.consumers

    @classmethod
    def from_propagation_group(cls, group: PropagationGroup) -> "PruningGroup":
        """
        Creates an object by taking all necessary information from PropagationGroup.
        """
        return cls(copy(group.block), group.get_producers(), group.get_consumers())


def get_pruning_groups(
    graph: NNCFGraph,
    pruning_operations_metatypes: PruningOperationsMetatypeRegistry,
    prune_operations_types: List[str],
    dump_dir: Optional[Path] = None,
) -> List[PruningGroup]:
    """
    Determines how nodes of the given types should be pruned: which nodes should be pruned together, along which
    dimension, how many sequent channels with which offset. It's done by initializing PropagationMask's on the
    operations with prunable parameters (producers of pruning) and propagating them through the execution graph.

    :param graph: nncf graph to initialize and propagate masks.
    :param pruning_operations_metatypes: registry with operation metatypes pruning algorithm is aware of, i.e.
        metatypes describing operations with common pruning mask application and propagation properties, e.g.
        IdentityMaskForwardOps unifies operations that propagate pruning masks as is (relu, swish etc.), whereas
        Convolution unifies different convolution operations (conv1d, conv2d, conv3d) which accepts some input masks and
        provide some output masks.
    :param prune_operations_types: types of operations with prunable parameters.
    :param dump_dir: path to the directory for dumping debug files.
    :return: list of groups with parameters of pruning.
    """
    # 1. Initialize masks for producing nodes
    all_nodes_to_prune: List[NNCFNode] = graph.get_nodes_by_types(prune_operations_types)
    roots = {}
    for node in all_nodes_to_prune:
        assert isinstance(node.layer_attributes, (LinearLayerAttributes, ConvolutionLayerAttributes))
        pruning_dim = node.layer_attributes.get_target_dim_for_compression()
        output_tensors_shapes = [x.tensor_shape for x in graph.get_output_edges(node)]
        assert not len(set(output_tensors_shapes)) > 1, node.node_name
        output_tensors_shape = output_tensors_shapes[0]
        # TODO: (107663) generalize by introducing get_output_dim_affected_by_compression for layer_attributes
        target_output_dim_for_compression = len(output_tensors_shape) - 1
        root_group = PropagationGroup(block=PruningBlock(), producers={ProducerInfo(node.node_id, pruning_dim)})
        mask = PropagationMask(dim_groups_map={target_output_dim_for_compression: [root_group]})
        roots[node.node_id] = root_group
        node.attributes["output_mask"] = mask

    def get_attributes_fn(node: NNCFNode) -> Dict[str, Any]:
        result = {"metatype": str(node.metatype.name), "node_id": str(node.node_id)}
        if node.layer_attributes:
            result.update(map(lambda pair: (pair[0], str(pair[1])), node.layer_attributes.__dict__.items()))
        if "output_mask" in node.attributes:
            output_mask = node.attributes["output_mask"]
            if output_mask:
                result["output_mask"] = str(output_mask)
        return result

    try:
        # 2. Propagate masks
        MaskPropagationAlgorithm(graph, pruning_operations_metatypes).mask_propagation()
    finally:
        block_hierarchy = BlockHierarchy(list(roots.values()))
        if dump_dir is not None:
            block_hierarchy.visualize_graph(dump_dir / "latest_block_group_hierarchy.dot")
            save_for_netron(graph, str(dump_dir / "latest_propagated_graph.xml"), get_attributes_fn=get_attributes_fn)

    propagation_groups = block_hierarchy.get_groups_on_leaves()
    not_invalid_groups = filter(lambda group: not group.is_invalid, propagation_groups)
    return [PruningGroup.from_propagation_group(pg) for pg in not_invalid_groups]


def select_largest_groups(pruning_groups: List[PruningGroup]) -> List[PruningGroup]:
    """
    Selects largest pruning groups with larger number of pruning blocks.
    """
    selected_producers = set()
    sorted_groups = sorted(pruning_groups, key=lambda group: len(group.producers), reverse=True)
    result = []
    for group in sorted_groups:
        producers = {producer_info.node_id for producer_info in group.producers}
        if not producers.intersection(selected_producers):
            selected_producers.update(producers)
            result.append(group)
    return result



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple

from nncf.common.graph import NNCFNode
from nncf.common.sparsity.statistics import ConstSparsityStatistics
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoBuilder
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoController
from nncf.torch.sparsity.collector import PTSparseModelStatisticsCollector
from nncf.torch.sparsity.layers import BinaryMask


@PT_COMPRESSION_ALGORITHMS.register("const_sparsity")
class ConstSparsityBuilder(BaseSparsityAlgoBuilder):
    def create_weight_sparsifying_operation(self, target_module_node: NNCFNode, compression_lr_multiplier: float):
        return BinaryMask(target_module_node.layer_attributes.get_weight_shape())

    def _build_controller(self, model: NNCFNetwork) -> PTCompressionAlgorithmController:
        return ConstSparsityController(model, self._sparsified_module_info)

    def _are_frozen_layers_allowed(self) -> Tuple[bool, str]:
        return True, "Frozen layers are allowed for const sparsity"


@api()
class ConstSparsityController(BaseSparsityAlgoController):
    """
    Controller for the auxiliary constant sparsity algorithm in PT.
    """

    def freeze(self):
        pass

    def set_sparsity_level(self, sparsity_level: float):
        pass

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        collector = PTSparseModelStatisticsCollector(self.model, self.sparsified_module_info)
        model_statistics = collector.collect()
        stats = ConstSparsityStatistics(model_statistics)

        nncf_stats = NNCFStatistics()
        nncf_stats.register("const_sparsity", stats)
        return nncf_stats



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC
from abc import abstractmethod
from enum import Enum
from functools import partial
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import distributed
from torch import nn

import nncf
from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.quantizer_setup import QuantizerSetupBase
from nncf.common.quantization.quantizers import calculate_asymmetric_level_ranges
from nncf.common.quantization.quantizers import calculate_symmetric_level_ranges
from nncf.common.quantization.quantizers import get_num_levels
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import QuantizerSpec
from nncf.common.utils.debug import is_debug
from nncf.common.utils.registry import Registry
from nncf.torch.checkpoint_loading import OPTIONAL_PARAMETERS_REGISTRY
from nncf.torch.dynamic_graph.context import no_nncf_trace
from nncf.torch.functions import clamp
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import TargetType
from nncf.torch.layer_utils import COMPRESSION_MODULES
from nncf.torch.layer_utils import CompressionParameter
from nncf.torch.layer_utils import StatefullModuleInterface
from nncf.torch.quantization.quantize_functions import ExportQuantizeToFakeQuantize
from nncf.torch.quantization.quantize_functions import ExportQuantizeToONNXQuantDequant
from nncf.torch.quantization.quantize_functions import TuneRange
from nncf.torch.quantization.quantize_functions import asymmetric_quantize
from nncf.torch.quantization.quantize_functions import decompress_asymmetric
from nncf.torch.quantization.quantize_functions import decompress_symmetric
from nncf.torch.quantization.quantize_functions import get_scale_zp_from_input_low_input_high
from nncf.torch.quantization.quantize_functions import pack_int4
from nncf.torch.quantization.quantize_functions import pack_uint4
from nncf.torch.quantization.quantize_functions import symmetric_quantize
from nncf.torch.quantization.quantize_functions import unpack_int4
from nncf.torch.quantization.quantize_functions import unpack_uint4
from nncf.torch.return_types import maybe_get_values_from_torch_return_type
from nncf.torch.return_types import maybe_wrap_to_torch_return_type
from nncf.torch.utils import get_flat_tensor_contents_string
from nncf.torch.utils import get_model_device
from nncf.torch.utils import is_tracing_state
from nncf.torch.utils import no_jit_trace

QUANTIZATION_MODULES = Registry("quantization_modules")
INITIALIZABLE_MODULES = Registry("initializable_modules")


class QuantizerExportMode(Enum):
    FAKE_QUANTIZE = "fake_quantize"
    ONNX_QUANTIZE_DEQUANTIZE_PAIRS = "quantize_dequantize"


class PTQSpecStateNames:
    NUM_BITS = "num_bits"
    MODE = "mode"
    SIGNED_TO_FORCE = "signedness_to_force"
    NARROW_RANGE = "narrow_range"
    HALF_RANGE = "half_range"
    SCALE_SHAPE = "scale_shape"
    LOGARITHM_SCALE = "logarithm_scale"
    IS_QUANTIZED_ON_EXPORT = "is_quantized_on_export"
    COMPRESSION_LR_MULTIPLIER = "compression_lr_multiplier"


class PTQuantizerSpec(QuantizerSpec):
    _state_names = PTQSpecStateNames

    def __init__(
        self,
        num_bits: int,
        mode: QuantizationMode,
        signedness_to_force: Optional[bool],
        narrow_range: bool,
        half_range: bool,
        scale_shape: Tuple[int, ...],
        logarithm_scale: bool,
        is_quantized_on_export: bool = False,
        compression_lr_multiplier: float = None,
    ):
        """
        :param scale_shape: Shape of quantizer scale parameters
        :param logarithm_scale: Whether to use log of scale as optimized parameter instead of scale itself.
        :param compression_lr_multiplier: Used to increase/decrease gradients for quantization parameters.
        :param is_quantized_on_export: Export to onnx weights quantized or non quantized. Should not be True for
            activation quantizers.
        """
        super().__init__(num_bits, mode, signedness_to_force, narrow_range, half_range)
        self.per_channel = scale_shape != (1,)
        self.scale_shape = scale_shape
        self.logarithm_scale = logarithm_scale
        self.compression_lr_multiplier = compression_lr_multiplier
        self.is_quantized_on_export = is_quantized_on_export

    @classmethod
    def from_config(
        cls,
        qconfig: QuantizerConfig,
        narrow_range: bool,
        half_range: bool,
        scale_shape: Tuple[int],
        logarithm_scale: bool,
        is_quantized_on_export: bool,
        compression_lr_multiplier: float,
    ) -> "PTQuantizerSpec":
        return cls(
            qconfig.num_bits,
            qconfig.mode,
            qconfig.signedness_to_force,
            narrow_range,
            half_range,
            scale_shape,
            logarithm_scale,
            is_quantized_on_export,
            compression_lr_multiplier,
        )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "PTQuantizationPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        kwargs = {
            cls._state_names.NUM_BITS: state["num_bits"],
            cls._state_names.MODE: state["mode"],
            cls._state_names.SIGNED_TO_FORCE: state["signedness_to_force"],
            cls._state_names.NARROW_RANGE: state["narrow_range"],
            cls._state_names.HALF_RANGE: state["half_range"],
            cls._state_names.SCALE_SHAPE: state["scale_shape"],
            cls._state_names.LOGARITHM_SCALE: state["logarithm_scale"],
            cls._state_names.IS_QUANTIZED_ON_EXPORT: state["is_quantized_on_export"],
            cls._state_names.COMPRESSION_LR_MULTIPLIER: state["compression_lr_multiplier"],
        }
        return cls(**kwargs)

    def get_state(self):
        return {
            self._state_names.NUM_BITS: self.num_bits,
            self._state_names.MODE: self.mode,
            self._state_names.SIGNED_TO_FORCE: self.signedness_to_force,
            self._state_names.NARROW_RANGE: self.narrow_range,
            self._state_names.HALF_RANGE: self.half_range,
            self._state_names.SCALE_SHAPE: self.scale_shape,
            self._state_names.LOGARITHM_SCALE: self.logarithm_scale,
            self._state_names.IS_QUANTIZED_ON_EXPORT: self.is_quantized_on_export,
            self._state_names.COMPRESSION_LR_MULTIPLIER: self.compression_lr_multiplier,
        }


class PTQPointStateNames:
    QSPEC = "qspec"
    TARGET_POINT = "target_point"
    NAMES_OF_QUANTIZED_OPS = "directly_quantized_operator_node_names"


class PTQuantizationPoint:
    _state_names = PTQPointStateNames

    def __init__(
        self,
        qspec: PTQuantizerSpec,
        target_point: PTTargetPoint,
        directly_quantized_operator_node_names: List[NNCFNodeName],
    ):
        self.qspec = qspec
        self.target_point = target_point
        self.directly_quantized_operator_node_names = directly_quantized_operator_node_names

    def is_activation_quantization_point(self) -> bool:
        return not self.is_weight_quantization_point()

    def is_weight_quantization_point(self) -> bool:
        return self.target_point.target_type == TargetType.OPERATION_WITH_WEIGHTS

    def __str__(self):
        return str(self.target_point) + " " + str(self.qspec)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.TARGET_POINT: self.target_point.get_state(),
            self._state_names.QSPEC: self.qspec.get_state(),
            self._state_names.NAMES_OF_QUANTIZED_OPS: self.directly_quantized_operator_node_names,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "PTQuantizationPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        kwargs = {
            cls._state_names.TARGET_POINT: PTTargetPoint.from_state(state[cls._state_names.TARGET_POINT]),
            cls._state_names.QSPEC: PTQuantizerSpec.from_state(state[cls._state_names.QSPEC]),
            cls._state_names.NAMES_OF_QUANTIZED_OPS: state[cls._state_names.NAMES_OF_QUANTIZED_OPS],
        }
        return cls(**kwargs)


class PTQSetupStateNames:
    SHARED_INPUT_OPERATION_SET_GROUPS = "shared_input_operation_set_groups"
    UNIFIED_SCALE_GROUPS = "unified_scale_groups"
    QUANTIZATION_POINTS = "quantization_points"


class PTQuantizerSetup(QuantizerSetupBase):
    _state_names = PTQSetupStateNames

    def __init__(self, unified_scale_groups, shared_input_operation_set_groups):
        super().__init__()
        self.unified_scale_groups = unified_scale_groups
        self.shared_input_operation_set_groups = shared_input_operation_set_groups

    @classmethod
    def from_state(cls, state: Dict) -> "PTQuantizerSetup":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """

        def decode_qp(pair):
            str_qp_id, qp_state = pair
            return int(str_qp_id), PTQuantizationPoint.from_state(qp_state)

        def list2set(pair):
            str_idx, qp_id_list = pair
            return int(str_idx), set(qp_id_list)

        unified_scale_groups = dict(map(list2set, state[cls._state_names.UNIFIED_SCALE_GROUPS].items()))
        shared_input_operation_set_groups_state = state[cls._state_names.SHARED_INPUT_OPERATION_SET_GROUPS]
        setup = PTQuantizerSetup(unified_scale_groups, shared_input_operation_set_groups_state)
        setup.quantization_points = dict(map(decode_qp, state[cls._state_names.QUANTIZATION_POINTS].items()))
        setup.shared_input_operation_set_groups = dict(map(list2set, shared_input_operation_set_groups_state.items()))
        return setup

    def get_state(self):
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """

        def set2list(pair):
            i, qp_id_set = pair
            return i, list(qp_id_set)

        quantization_points_state = {qp_id: qp.get_state() for qp_id, qp in self.quantization_points.items()}
        unified_scale_groups_state = dict(map(set2list, self.unified_scale_groups.items()))
        shared_input_operation_set_groups_state = dict(map(set2list, self.shared_input_operation_set_groups.items()))
        return {
            self._state_names.QUANTIZATION_POINTS: quantization_points_state,
            self._state_names.UNIFIED_SCALE_GROUPS: unified_scale_groups_state,
            self._state_names.SHARED_INPUT_OPERATION_SET_GROUPS: shared_input_operation_set_groups_state,
        }

    def add_quantization_point(self, qp_id: QuantizationPointId, qp: PTQuantizationPoint):
        self.quantization_points[qp_id] = qp


class BaseQuantizer(nn.Module, StatefullModuleInterface, ABC):
    def __init__(self, qspec: PTQuantizerSpec):
        super().__init__()
        self._qspec = qspec
        self._narrow_range = qspec.narrow_range
        self._signedness_to_force = qspec.signedness_to_force
        self._is_using_log_scale_storage = qspec.logarithm_scale
        self._half_range = qspec.half_range
        self._is_quantized_on_export = qspec.is_quantized_on_export
        self._num_bits = CompressionParameter(
            torch.IntTensor([qspec.num_bits]),
            requires_grad=False,
            compression_lr_multiplier=qspec.compression_lr_multiplier,
        )
        OPTIONAL_PARAMETERS_REGISTRY.register("_num_bits")

        # These must be made buffers, since they impact the "forward" behaviour and the model can be used
        # in DDP scenarios, so these must be properly synchronized across processes.
        self.register_buffer("_level_low", torch.IntTensor([0]), persistent=False)
        self.register_buffer("_level_high", torch.IntTensor([0]), persistent=False)

        ENABLED_VAR_NAME = "enabled"
        self.register_buffer(ENABLED_VAR_NAME, torch.IntTensor([1]))
        OPTIONAL_PARAMETERS_REGISTRY.register(ENABLED_VAR_NAME)
        self.initialized = False
        self.call_count = 0
        self._scale_shape = qspec.scale_shape
        self._export_mode = QuantizerExportMode.FAKE_QUANTIZE

        class LoadStateListener:
            """
            Check whether a quantization module are going to be updated by new values from state_dict or checkpoint.
            """

            def __init__(self, module):
                self.hook = module._register_load_state_dict_pre_hook(partial(self.hook_fn, module=module))

            def hook_fn(
                self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs, module
            ):
                for module_key in module.state_dict():
                    candidate = prefix + module_key
                    if candidate in state_dict:
                        module.initialized = True

            def close(self):
                self.hook.remove()

        self.load_listener = LoadStateListener(self)

    @property
    def level_low(self) -> int:
        return self._level_low.item()

    @level_low.setter
    def level_low(self, val: int):
        self._level_low.fill_(val)

    @property
    def level_high(self) -> int:
        return self._level_high.item()

    @level_high.setter
    def level_high(self, val: int):
        self._level_high.fill_(val)

    @property
    def levels(self):
        return get_num_levels(self.level_low, self.level_high)

    @abstractmethod
    def enable_gradients(self):
        pass

    @abstractmethod
    def disable_gradients(self):
        pass

    def is_enabled_quantization(self):
        with no_jit_trace():
            return self.enabled[0].item() == 1

    def enable_quantization(self):
        self.enabled[0] = 1
        self.enable_gradients()

    def disable_quantization(self):
        self.enabled[0] = 0
        self.disable_gradients()

    def forward(self, x: Union[torch.Tensor, tuple]):
        """
        Method that unwraps return types if it is needed
        before acutal quantization forward impl
        """
        x_unwrapped = maybe_get_values_from_torch_return_type(x)
        result = self._forward_impl(x_unwrapped)
        return maybe_wrap_to_torch_return_type(result, x)

    def _forward_impl(self, x: torch.Tensor):
        if is_debug():
            self.call_count += 1
        # TODO: refactor to get rid of extra if's and calls on each forward
        if not self.is_enabled_quantization():
            return x
        is_exporting = is_tracing_state()
        if is_exporting:
            with no_nncf_trace():
                x = self.run_export_quantization(x)

            # The underlying operator (registered via register_operator) must be executed,
            # otherwise the dynamic graph won't be traced as it was during regular inference.
            # While this does not impact the regular, non-RNN models, for which the graph
            # building and pre-/post-hook calling is only determined by input-agnostic,
            # graph-structure independent trace info (e.g. current op scope and call count),
            # this is important for LSTMs etc. where determining the "first nodes in iteration
            # scopes" depends on whether the input tensors to an operation were traced or not.
            return self.quantize(x, execute_traced_op_as_identity=True)

        return self.quantize(x, execute_traced_op_as_identity=False)

    @abstractmethod
    def quantize(self, x, execute_traced_op_as_identity: bool = False):
        pass

    def reset_call_counter(self):
        self.call_count = 0

    @abstractmethod
    def get_trainable_params(self) -> Dict[str, torch.Tensor]:
        pass

    def apply_minmax_init(self, min_values: torch.Tensor, max_values: torch.Tensor, log_module_name: str = None):
        """min_values and max_values must have the same shape as specified in self.scale_shape"""
        if self.initialized:
            nncf_logger.debug(f"Skipped initializing {log_module_name} - loaded from checkpoint")
            return

        if torch.all(torch.isinf(min_values)) or torch.all(torch.isinf(max_values)):
            raise ValueError(f"Statistics are not collected for {log_module_name}")

        if torch.any(torch.eq(min_values, np.inf)) or torch.any(torch.eq(max_values, -np.inf)):
            raise ValueError(f"Some of the values in statistics have infinite value for {log_module_name}")

        own_device = get_model_device(self)
        min_values = min_values.to(own_device)
        max_values = max_values.to(own_device)
        self._apply_minmax_init(min_values, max_values, log_module_name)

    @abstractmethod
    def _apply_minmax_init(self, min_values: torch.Tensor, max_values: torch.Tensor, log_module_name: str = None):
        pass

    @abstractmethod
    def set_levels(self):
        """Must set the self._level_low and self._level_high buffers according to the current quantizer state
        and type, and called whenever the state of the quantizer is updated in a way that affects the effective level
        ranges."""

    @property
    def is_half_range(self):
        return self._half_range

    @property
    def is_using_log_scale_storage(self):
        return self._is_using_log_scale_storage

    @property
    @abstractmethod
    def signed(self):
        pass

    @property
    def num_bits(self):
        with no_jit_trace():
            return self._num_bits.item()

    @num_bits.setter
    def num_bits(self, num_bits: int):
        self._num_bits.fill_(num_bits)
        self.set_levels()

    @property
    def narrow_range(self) -> bool:
        return self._narrow_range

    @property
    def scale_shape(self) -> Tuple[int, ...]:
        # Per-tensor scale shapes are (1,)
        return self._scale_shape

    def broadcast_initialized_params(self, src: int = 0):
        distributed.broadcast(self._num_bits, src=src)

    def set_export_mode(self, mode: QuantizerExportMode):
        self._export_mode = mode

    @abstractmethod
    def _get_input_low_input_high(self):
        pass

    @abstractmethod
    def _prepare_export_quantization(self, x: torch.Tensor):
        pass

    def _prepare_fq_export_quantization(self, x: torch.Tensor):
        x, level_high, level_low, input_low, input_high = self._prepare_export_quantization(x)
        with no_jit_trace():
            levels = level_high - level_low + 1
        return x, levels, input_low, input_high

    def _prepare_qdq_export_quantization(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, int]:
        x, level_high, level_low, input_low, input_high = self._prepare_export_quantization(x)
        with no_jit_trace():
            levels = level_high - level_low + 1
            assert levels in [255, 256], "Can only export to INT8 256-level ONNX Quantize/Dequantize pairs"

            y_scale, y_zero_point = get_scale_zp_from_input_low_input_high(level_low, level_high, input_low, input_high)
            possible_axes = self._possible_per_channel_dimensions()
            if len(possible_axes) > 1:
                raise nncf.InternalError(
                    f"Impossible to determine the per-channel axis for a scale shape {self.scale_shape} - "
                    f"more than one dimension is >1"
                )
            if not possible_axes:
                # Impossible to determine proper axis for per-channel quantization because we have
                # scale shape ~ [1, 1, 1, 1], therefore falling back to per-tensor style export
                axis = 1  # default value by opset, ignored in per-tensor quantization anyway
                y_scale = y_scale.flatten()[0]
                y_zero_point = y_zero_point.flatten()[0]
            else:
                axis = possible_axes[0]

        return x, y_scale, y_zero_point, axis

    def _possible_per_channel_dimensions(self) -> List[int]:
        return [i for i in range(len(self.scale_shape)) if self.scale_shape[i] > 1]

    def run_export_quantization(self, x: torch.Tensor):
        with torch.no_grad():
            if self._export_mode == QuantizerExportMode.FAKE_QUANTIZE:
                x, levels, input_low, input_high = self._prepare_fq_export_quantization(x)
                q_min, q_max, scale, zero_point = self.get_parameters_for_torch_fq()

                ch_axis = np.argmax(self.scale_shape) if self.per_channel else None

                return ExportQuantizeToFakeQuantize.apply(
                    x, levels, input_low, input_high, input_low, input_high, scale, zero_point, q_min, q_max, ch_axis
                )
            if self._export_mode == QuantizerExportMode.ONNX_QUANTIZE_DEQUANTIZE_PAIRS:
                x, y_scale, y_zero_point, axis = self._prepare_qdq_export_quantization(x)
                return ExportQuantizeToONNXQuantDequant.apply(x, y_scale, y_zero_point, axis)
        raise nncf.InternalError("Unknown export mode")

    def extra_repr(self):
        return "bit={}, ch={}".format(self.num_bits, self.per_channel)

    @abstractmethod
    def get_quantizer_config(self) -> QuantizerConfig:
        pass

    @property
    def per_channel(self) -> bool:
        numel = 1
        for el in self.scale_shape:
            numel *= el
        is_per_tensor = (numel == 1) and (len(self.scale_shape) == 1)
        return not is_per_tensor

    @abstractmethod
    def get_parameters_for_torch_fq(self) -> Tuple[int, int, torch.Tensor, torch.Tensor]:
        """
        Get parameters for conversion to native FakeQuantize.

        :return: A Tuple
            quant_max - Fixed the low quant number.
            quant_min - Fixed the high quant number.
            scale - Quantizer scale.
            zero_point - Quantizer zero point.
        """

    def get_config(self):
        return self._qspec.get_state()

    @classmethod
    def from_config(cls, state) -> "BaseQuantizer":
        qsetup = PTQuantizerSpec.from_state(state)
        return cls(qsetup)


class QuantizersSwitcher:
    """Enables/disables quantizers with saving and restoring original state"""

    def __init__(self, quantizers: List[BaseQuantizer]):
        self.originally_disabled: List[BaseQuantizer] = []
        self.originally_enabled: List[BaseQuantizer] = []
        self._quantizers = quantizers

    def disable_quantizers(self):
        for module in self._quantizers:
            if not module.is_enabled_quantization():
                self.originally_disabled.append(module)
            if module not in self.originally_enabled:
                module.disable_quantization()
        self.originally_enabled = []

    def enable_quantizers(self):
        for module in self._quantizers:
            if module.is_enabled_quantization():
                self.originally_enabled.append(module)
            if module not in self.originally_disabled:
                module.enable_quantization()
        self.originally_disabled = []


class StorageRedirectingLoadStateDictHook:
    def __init__(
        self, storage_attribute_in_module: str, name_in_state_dict: str, use_log_storage_in_module: bool = False
    ):
        self._storage_attribute_in_module = storage_attribute_in_module
        self._name_in_state_dict = name_in_state_dict
        self._use_log_storage_in_module = use_log_storage_in_module

    def __call__(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs) -> None:
        state_dict_key = prefix + self._name_in_state_dict
        if state_dict_key in state_dict:
            v = state_dict.pop(state_dict_key)
            if self._use_log_storage_in_module:
                v = v.abs().log().detach()
            state_dict[prefix + self._storage_attribute_in_module] = v
        else:
            missing_keys.append(state_dict_key)


class StorageRedirectingStateDictHook:
    def __init__(
        self, storage_attribute_in_module: str, name_in_state_dict: str, use_log_storage_in_module: bool = False
    ):
        self._storage_attribute_in_module = storage_attribute_in_module
        self._name_in_state_dict = name_in_state_dict
        self._use_log_storage_in_module = use_log_storage_in_module

    def __call__(self, module, state_dict, prefix, local_metadata) -> None:
        v = state_dict.pop(prefix + self._storage_attribute_in_module)
        if self._use_log_storage_in_module:
            v = v.exp().detach()
        state_dict[prefix + self._name_in_state_dict] = v


@COMPRESSION_MODULES.register()
@QUANTIZATION_MODULES.register(QuantizationMode.SYMMETRIC)
class SymmetricQuantizer(BaseQuantizer):
    SCALE_PARAM_NAME = "scale"
    _SCALE_PARAM_STORAGE_ATTR = "_scale_param_storage"

    def __init__(self, qspec: PTQuantizerSpec):
        super().__init__(qspec)
        self.signed_tensor = CompressionParameter(
            torch.IntTensor([0]), requires_grad=False, compression_lr_multiplier=qspec.compression_lr_multiplier
        )
        self.collect_scale_statistics = False

        setattr(
            self,
            self._SCALE_PARAM_STORAGE_ATTR,
            CompressionParameter(
                torch.ones(self.scale_shape),
                requires_grad=True,
                compression_lr_multiplier=qspec.compression_lr_multiplier,
            ),
        )
        if self._is_using_log_scale_storage:
            self._scale_param_storage.data.log_()
            self.eps = 0
        else:
            self.eps = 1e-16
        if qspec.signedness_to_force is not None:
            self.signed = bool(qspec.signedness_to_force)
        self.set_levels()

        self._register_load_state_dict_pre_hook(
            StorageRedirectingLoadStateDictHook(
                storage_attribute_in_module=self._SCALE_PARAM_STORAGE_ATTR,
                name_in_state_dict=self.SCALE_PARAM_NAME,
                use_log_storage_in_module=self._is_using_log_scale_storage,
            )
        )

        self._register_state_dict_hook(
            StorageRedirectingStateDictHook(
                storage_attribute_in_module=self._SCALE_PARAM_STORAGE_ATTR,
                name_in_state_dict=self.SCALE_PARAM_NAME,
                use_log_storage_in_module=self._is_using_log_scale_storage,
            )
        )

        # Values of level_low, level_high must be recalculated for load new signed parameter.
        self.register_load_state_dict_post_hook(lambda module, _: module.set_levels())

    @property
    def scale(self):
        return self._scale_param_storage.exp() if self._is_using_log_scale_storage else self._scale_param_storage

    @scale.setter
    def scale(self, v):
        self._scale_param_storage = v
        if self._is_using_log_scale_storage:
            self._scale_param_storage.data.log_()

    def __setattr__(self, key, value):
        """
        Need to handle the redirect-storage attributes (which are implemented using Python properties
        here) specially - otherwise the torch.nn.Module's __setattr__ will try to set them during
        assignment.
        """
        if key == self.SCALE_PARAM_NAME:
            object.__setattr__(self, key, value)
        else:
            super().__setattr__(key, value)

    def enable_gradients(self):
        self._scale_param_storage.requires_grad = True

    def disable_gradients(self):
        self._scale_param_storage.requires_grad = False

    def set_levels(self):
        scaled_num_bits = 1 if self._half_range else 0
        self.level_low, self.level_high = calculate_symmetric_level_ranges(
            self.num_bits - scaled_num_bits, self.signed, self._narrow_range
        )

    @property
    def signed(self):
        with no_jit_trace():
            return self.signed_tensor.item() == 1

    @signed.setter
    def signed(self, signed: bool):
        self.signed_tensor.fill_(int(signed))
        self.set_levels()

    def quantize(self, x, execute_traced_op_as_identity: bool = False):
        return symmetric_quantize(
            x, self.levels, self.level_low, self.level_high, self.scale, self.eps, skip=execute_traced_op_as_identity
        )

    def get_trainable_params(self) -> Dict[str, torch.Tensor]:
        return {self.SCALE_PARAM_NAME: self.scale.detach()}

    def _apply_minmax_init(self, min_values, max_values, log_module_name: str = None):
        sign = torch.any(torch.lt(min_values, 0))
        if self._signedness_to_force is not None and sign != self._signedness_to_force:
            nncf_logger.debug(f"Forcing signed to {self._signedness_to_force} for module {log_module_name}")
            sign = self._signedness_to_force
        self.signed = sign

        abs_max = torch.max(torch.abs(max_values), torch.abs(min_values))
        SCALE_LOWER_THRESHOLD = 0.1
        mask = torch.gt(abs_max, SCALE_LOWER_THRESHOLD)
        self._scale_param_storage.data = torch.where(
            mask, abs_max, SCALE_LOWER_THRESHOLD * torch.ones_like(self._scale_param_storage)
        )
        if self._is_using_log_scale_storage:
            self._scale_param_storage.data.log_()

        nncf_logger.debug(
            f"Set sign: {self.signed} and scale: {get_flat_tensor_contents_string(self.scale)} for {log_module_name}"
        )

    def broadcast_initialized_params(self, src: int = 0):
        super().broadcast_initialized_params(src)
        distributed.broadcast(self._scale_param_storage, src=src)
        distributed.broadcast(self.signed_tensor, src=src)

    def get_input_low_input_high(self):
        return self._get_input_low_input_high(self.scale, self.level_low, self.level_high, self.eps)

    def _get_input_low_input_high(self, scale, level_low, level_high, eps):
        input_range = abs(scale) + eps
        input_low = input_range * level_low / level_high
        input_high = input_range
        return input_low, input_high

    def _prepare_export_quantization(self, x: torch.Tensor):
        with no_jit_trace():
            input_low, input_high = self._get_input_low_input_high(
                self.scale, self.level_low, self.level_high, self.eps
            )
            level_low = self.level_low
            level_high = self.level_high
            if self._half_range:
                x = torch.min(torch.max(x, input_low), input_high)
                level_low = 2 * self.level_low
                level_high = 2 * self.level_high + 1
                input_low, input_high = self._get_input_low_input_high(
                    level_high / self.level_high * self.scale, level_low, level_high, self.eps
                )
            if self._is_quantized_on_export:
                x = self.quantize(x, execute_traced_op_as_identity=False)
        return x, level_high, level_low, input_low, input_high

    def get_parameters_for_torch_fq(self) -> Tuple[int, int, torch.Tensor, torch.Tensor]:
        """
        Get parameters for conversion to native FakeQuantize.

        :return: A Tuple
            quant_max - Fixed the low quant number.
            quant_min - Fixed the high quant number.
            scale - Quantizer scale.
            zero_point - Quantizer zero point.
        """
        with torch.no_grad(), no_jit_trace():
            input_low, input_high = self._get_input_low_input_high(
                self.scale, self.level_low, self.level_high, self.eps
            )
            level_low = self.level_low
            level_high = self.level_high

            scale, zero_point = get_scale_zp_from_input_low_input_high(level_low, level_high, input_low, input_high)

            if self.narrow_range:
                if level_low < 0:
                    level_low -= 1
                else:
                    level_high += 1

            if self._half_range:
                level_low = 2 * level_low
                level_high = 2 * level_high + 1

            scale = scale.view(-1)
            zero_point = zero_point.view(-1).to(dtype=torch.int32)

        return level_low, level_high, scale, zero_point

    def get_quantizer_config(self) -> QuantizerConfig:
        return QuantizerConfig(
            num_bits=self.num_bits,
            mode=QuantizationMode.SYMMETRIC,
            signedness_to_force=self.signed,
            per_channel=self.per_channel,
        )


@COMPRESSION_MODULES.register()
@QUANTIZATION_MODULES.register(QuantizationMode.ASYMMETRIC)
class AsymmetricQuantizer(BaseQuantizer):
    INPUT_LOW_PARAM_NAME = "input_low"
    INPUT_RANGE_PARAM_NAME = "input_range"
    _INPUT_RANGE_PARAM_STORAGE_ATTR = "_input_range_param_storage"

    def __init__(self, qspec: PTQuantizerSpec):
        super().__init__(qspec)
        self.input_low = CompressionParameter(
            torch.zeros(self.scale_shape), requires_grad=True, compression_lr_multiplier=qspec.compression_lr_multiplier
        )
        setattr(
            self,
            self._INPUT_RANGE_PARAM_STORAGE_ATTR,
            CompressionParameter(
                torch.ones(self.scale_shape),
                requires_grad=True,
                compression_lr_multiplier=qspec.compression_lr_multiplier,
            ),
        )

        if self._is_using_log_scale_storage:
            self._input_range_param_storage.data.log_()
            self.eps = 0
        else:
            self.eps = 1e-16
        self.set_levels()

        self._register_load_state_dict_pre_hook(
            StorageRedirectingLoadStateDictHook(
                storage_attribute_in_module=self._INPUT_RANGE_PARAM_STORAGE_ATTR,
                name_in_state_dict=self.INPUT_RANGE_PARAM_NAME,
                use_log_storage_in_module=self._is_using_log_scale_storage,
            )
        )

        self._register_state_dict_hook(
            StorageRedirectingStateDictHook(
                storage_attribute_in_module=self._INPUT_RANGE_PARAM_STORAGE_ATTR,
                name_in_state_dict=self.INPUT_RANGE_PARAM_NAME,
                use_log_storage_in_module=self._is_using_log_scale_storage,
            )
        )

    @property
    def input_range(self):
        if self._is_using_log_scale_storage:
            return self._input_range_param_storage.exp()
        return self._input_range_param_storage

    @input_range.setter
    def input_range(self, v: torch.Tensor):
        self._input_range_param_storage = v
        if self._is_using_log_scale_storage:
            self._input_range_param_storage.data.log_()

    def __setattr__(self, key, value):
        """
        Need to handle the redirect-storage attributes (which are implemented using Python properties
        here) specially - otherwise the torch.nn.Module's __setattr__ will try to set them during
        assignment.
        """
        if key == self.INPUT_RANGE_PARAM_NAME:
            object.__setattr__(self, key, value)
        else:
            super().__setattr__(key, value)

    def enable_gradients(self):
        self.input_low.requires_grad = True
        self._input_range_param_storage.requires_grad = True

    def disable_gradients(self):
        self.input_low.requires_grad = False
        self._input_range_param_storage.requires_grad = False

    @property
    def signed(self):
        return True

    def set_levels(self):
        scaled_num_bits = 1 if self._half_range else 0
        self.level_low, self.level_high = calculate_asymmetric_level_ranges(self.num_bits - scaled_num_bits)

    def quantize(self, x, execute_traced_op_as_identity: bool = False):
        return asymmetric_quantize(
            x,
            self.levels,
            self.level_low,
            self.level_high,
            self.input_low,
            self.input_range,
            self.eps,
            skip=execute_traced_op_as_identity,
        )

    def get_trainable_params(self) -> Dict[str, torch.Tensor]:
        return {
            self.INPUT_LOW_PARAM_NAME: self.input_low.detach(),
            self.INPUT_RANGE_PARAM_NAME: self.input_range.detach(),
        }

    def _apply_minmax_init(self, min_values, max_values, log_module_name: str = None):
        ranges = max_values - min_values
        max_range = torch.max(max_values - min_values)
        eps = 1e-2
        correction = (clamp(ranges, low=eps * max_range, high=max_range) - ranges) * 0.5
        self._input_range_param_storage.data = (ranges + 2 * correction).data
        if self._is_using_log_scale_storage:
            self._input_range_param_storage.data.log_()

        self.input_low.data = (min_values - correction).data

        nncf_logger.debug(
            f"Set input_low: {get_flat_tensor_contents_string(self.input_low)} "
            f"and input_range: {get_flat_tensor_contents_string(self.input_range)} for {log_module_name}"
        )

    def broadcast_initialized_params(self, src: int = 0):
        super().broadcast_initialized_params(src)
        distributed.broadcast(self.input_low, src)
        distributed.broadcast(self._input_range_param_storage, src)

    def get_input_low_input_high(self):
        return self._get_input_low_input_high(self.input_range, self.input_low, self.levels, self.eps)

    def _get_input_low_input_high(self, input_range, input_low, levels, eps):
        input_range_safe = abs(input_range) + eps
        input_low, input_range_tuned = TuneRange.apply(input_low, input_range_safe, levels)
        input_high = input_low + input_range_tuned
        return input_low, input_high

    def _prepare_export_quantization(self, x: torch.Tensor):
        with no_jit_trace():
            input_low, input_high = self._get_input_low_input_high(
                self.input_range, self.input_low, self.levels, self.eps
            )
            level_low = self.level_low
            level_high = self.level_high
            if self._half_range:
                x = torch.min(torch.max(x, input_low), input_high)
                level_low = 2 * level_low
                level_high = 2 * level_high + 1
                input_low, input_high = self._get_input_low_input_high(
                    level_high / self.level_high * self.input_range, self.input_low, self.levels, self.eps
                )
            if self._is_quantized_on_export:
                x = self.quantize(x, execute_traced_op_as_identity=False)
        return x, level_high, level_low, input_low, input_high

    def get_parameters_for_torch_fq(self) -> Tuple[int, int, torch.Tensor, torch.Tensor]:
        """
        Get parameters for conversion to native FakeQuantize.

        :return: A Tuple
            quant_max - Fixed the low quant number.
            quant_min - Fixed the high quant number.
            scale - Quantizer scale.
            zero_point - Quantizer zero point.
        """
        with torch.no_grad(), no_jit_trace():
            input_low, input_high = self._get_input_low_input_high(
                self.input_range, self.input_low, self.levels, self.eps
            )
            level_low = self.level_low
            level_high = self.level_high

            scale, zero_point = get_scale_zp_from_input_low_input_high(level_low, level_high, input_low, input_high)

            if self._half_range:
                level_low = 2 * level_low
                level_high = 2 * level_high + 1

            scale = scale.view(-1)
            zero_point = zero_point.view(-1).to(dtype=torch.int32)

        return level_low, level_high, scale, zero_point

    def get_quantizer_config(self) -> QuantizerConfig:
        return QuantizerConfig(
            num_bits=self.num_bits,
            mode=QuantizationMode.ASYMMETRIC,
            signedness_to_force=self.signed,
            per_channel=self.per_channel,
        )


def get_per_channel_scale_shape(input_shape, is_weights, channel_idx: int = None):
    scale_shape = [1 for _ in input_shape]
    if channel_idx is None:
        if is_weights:
            channel_idx = 0  # Per weight channel scales
        else:
            channel_idx = 1  # Per activation channel scales
    scale_shape[channel_idx] = input_shape[channel_idx]
    return scale_shape


def get_scale_shape(input_shape: List[int], is_weights: bool, per_channel: bool, channel_idx: int = None) -> List[int]:
    """
    Assumes that input_shape is supplied in either [B, C, H, W] or [N_out, N_in, H, W] format,
    or derivatives.
    :param input_shape: The input shape of the tensor; semantic meaning of dimensions should be as described above.
    :param is_weights: Whether the tensor corresponds to weights, in which case the per-channel scaling dimension
        is selected based on the [N_out, N_in, H, W] format
    :param per_channel: If True, will generate a per-channel scale shape, otherwise will generate a scale shape
        corresponding to per-tensor quantization
    :param channel_idx: The index of the per-channel dimension among input_shape.
    :return: The shape for the quantizer's scale tensors.
    """
    if not per_channel:
        return [1]
    return get_per_channel_scale_shape(input_shape, is_weights, channel_idx)


class BaseWeightsDecompressor(nn.Module, ABC):
    """
    Base class for implementing weights decompression modules within NNCF.

    This class is intended to serve as the foundation for modules that handle the decompression
    of quantized model weights. It provides an interface for defining the quantization mode and
    packing the weights according to the specified quantization strategy. Classes inheriting from
    this base class must implement the abstract methods for packing and handling the quantization mode.
    """

    @property
    @abstractmethod
    def quantization_mode(self) -> QuantizationMode:
        """
        Property that specifies the quantization mode used for compressing weights.

        This method must be implemented to return the specific mode of quantization that
        the decompressor is using, such as symmetric or asymmetric quantization.

        :return: The quantization mode as an instance of `QuantizationMode`.
        """

    @abstractmethod
    def pack_weight(self, weight: torch.Tensor) -> torch.Tensor:
        """
        Pack the given weight tensor according to the selected quantization mode.

        :param weight: The tensor containing the weight values to be packed.
        :return: The packed tensor.
        """


class INT8AsymmetricWeightsDecompressor(BaseWeightsDecompressor):
    """
    Applies asymmetric decompression of compressed weights in the forward pass
    """

    def __init__(self, scale: torch.Tensor, zero_point: torch.Tensor, result_dtype: Optional[torch.dtype] = None):
        """
        :param scale: A scale in quantization scheme
        :param zero_point: A zero point in quantization scheme
        :param result_dtype: (Optional) A data type that result should be cast to
        """
        super().__init__()
        self.register_buffer("_scale", scale.type(dtype=torch.float16))
        self.register_buffer("_zero_point", self.pack_weight(zero_point))
        self.result_dtype = result_dtype

    @property
    def quantization_mode(self) -> QuantizationMode:
        return QuantizationMode.ASYMMETRIC

    def pack_weight(self, weight: torch.Tensor) -> torch.Tensor:
        if torch.is_floating_point(weight):
            raise ValueError(f"Invalid weight dtype {weight.type}. Integer types are supported.")
        if torch.any((weight < 0) | (weight > 255)):
            raise ValueError("Weight values are not in [0, 255].")
        return weight.type(dtype=torch.uint8)

    def forward(self, x) -> torch.Tensor:
        result = decompress_asymmetric(x, self._scale, self._zero_point)
        result = result.type(dtype=self.result_dtype) if self.result_dtype is not None else result
        return result


class INT8SymmetricWeightsDecompressor(BaseWeightsDecompressor):
    """
    Applies symmetric decompression of compressed weights in the forward pass
    """

    def __init__(self, scale: torch.Tensor, result_dtype: Optional[torch.dtype] = None):
        """
        :param scale: A scale in quantization scheme
        :param result_dtype: (Optional) A data type that result should be cast to
        """
        super().__init__()
        self.register_buffer("_scale", scale.type(dtype=torch.float16))
        self.result_dtype = result_dtype

    @property
    def quantization_mode(self) -> QuantizationMode:
        return QuantizationMode.SYMMETRIC

    def pack_weight(self, weight: torch.Tensor) -> torch.Tensor:
        if torch.is_floating_point(weight):
            raise ValueError(f"Invalid weight dtype {weight.type}. Integer types are supported.")
        if torch.any((weight < -128) | (weight > 127)):
            raise ValueError("Weight values are not in [-128, 127].")
        return weight.type(dtype=torch.int8)

    def forward(self, x) -> torch.Tensor:
        result = decompress_symmetric(x, self._scale)
        result = result.type(dtype=self.result_dtype) if self.result_dtype is not None else result
        return result


class INT4AsymmetricWeightsDecompressor(BaseWeightsDecompressor):
    def __init__(
        self,
        scale: torch.Tensor,
        zero_point: torch.Tensor,
        compressed_weight_shape: Tuple[int, ...],
        result_shape: Optional[Tuple[int, ...]] = None,
        result_dtype: Optional[torch.dtype] = None,
    ):
        """
        :param scale: A scale in quantization scheme
        :param zero_point: A zero point in quantization scheme
        :param compressed_weight_shape: A compressed weight shape
        :param result_shape: (Optional) A shape that result should be reshaped
        :param result_dtype: (Optional) A data type that result should be cast to
        """
        super().__init__()
        self.register_buffer("_scale", scale.type(dtype=torch.float16))

        self.zero_point_shape = zero_point.shape
        self.register_buffer("_zero_point", self.pack_weight(zero_point))

        self.compressed_weight_shape = compressed_weight_shape
        self.result_shape = result_shape
        self.result_dtype = result_dtype

    @property
    def quantization_mode(self) -> QuantizationMode:
        return QuantizationMode.ASYMMETRIC

    def pack_weight(self, weight: torch.Tensor) -> torch.Tensor:
        if torch.is_floating_point(weight):
            raise ValueError(f"Invalid weight dtype {weight.type}. Integer types are supported.")
        if torch.any((weight < 0) | (weight > 15)):
            raise ValueError("Weight values are not in [0, 15].")
        return pack_uint4(weight.type(dtype=torch.uint8))

    def forward(self, x):
        x = unpack_uint4(x)
        x = x.reshape(self.compressed_weight_shape)

        zero_point = unpack_uint4(self._zero_point)
        zero_point = zero_point.reshape(self.zero_point_shape)

        result = decompress_asymmetric(x, self._scale, zero_point)
        result = result.reshape(self.result_shape) if self.result_shape is not None else result
        result = result.type(dtype=self.result_dtype) if self.result_dtype is not None else result
        return result


class INT4SymmetricWeightsDecompressor(BaseWeightsDecompressor):
    def __init__(
        self,
        scale: torch.Tensor,
        compressed_weight_shape: Tuple[int, ...],
        result_shape: Optional[Tuple[int, ...]] = None,
        result_dtype: Optional[torch.dtype] = None,
    ):
        """
        :param scale: A scale in quantization scheme
        :param compressed_weight_shape: A compressed weight shape
        :param result_shape: (Optional) A shape that result should be reshaped
        :param result_dtype: (Optional) A data type that result should be cast to
        """
        super().__init__()
        self.register_buffer("_scale", scale.type(dtype=torch.float16))

        self.compressed_weight_shape = compressed_weight_shape
        self.result_shape = result_shape
        self.result_dtype = result_dtype

    @property
    def quantization_mode(self) -> QuantizationMode:
        return QuantizationMode.SYMMETRIC

    def pack_weight(self, weight: torch.Tensor) -> torch.Tensor:
        if torch.is_floating_point(weight):
            raise ValueError(f"Invalid weight dtype {weight.type}. Integer types are supported.")
        if torch.any((weight < -8) | (weight > 7)):
            raise ValueError("Tensor values are not in [-8, 7].")
        return pack_int4(weight.type(dtype=torch.int8))

    def forward(self, x):
        x = unpack_int4(x)
        x = x.reshape(self.compressed_weight_shape)

        result = decompress_symmetric(x, self._scale)
        result = result.reshape(self.result_shape) if self.result_shape is not None else result
        result = result.type(dtype=self.result_dtype) if self.result_dtype is not None else result
        return result



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path
from typing import Any, Dict

from nncf.common.utils.backend import BackendType


def validate_backend(metadata: Dict[str, Any], backend: BackendType) -> None:
    """
    Checks whether backend in metadata is equal to a provided backend.

    :param data: Loaded statistics.
    :param backend: Provided backend.
    """
    if "backend" not in metadata:
        raise ValueError("The provided metadata has no information about backend.")
    data_backend = metadata["backend"]
    if data_backend != backend.value:
        raise ValueError(
            f"Backend in loaded statistics {data_backend} does not match the expected backend {backend.value}."
        )


def validate_statistics_files_exist(metadata: Dict[str, Any], dir_path: Path) -> None:
    """
    Checks whether all statistics files exist.

    :param metadata: Loaded metadata.
    :param dir_path: Path to the cache directory.
    """
    for file_name in metadata["mapping"]:
        file_path = dir_path / file_name
        if not file_path.exists():
            raise FileNotFoundError(f"One of the statistics file: {file_path} does not exist.")


def validate_cache(metadata: Dict[str, Any], dir_path: Path, backend: BackendType) -> None:
    """
    Validates cache directory.

    :param metadata: Metadata.
    :param dir_path: Path to the cache directory.
    :param backend: Backend type.
    """
    validate_backend(metadata, backend)
    validate_statistics_files_exist(metadata, dir_path)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import weakref
from typing import Any, Optional

from torch import nn


class RemovableHookHandle:
    """
    A handle to manage removable hooks that are stored in a dictionary within a PyTorch ModuleDict.
    The handle allows safe removal of specific hooks based on their unique identifiers.

    :param storage_ref: A weak reference to the storage (an nn.ModuleDict) containing the hooks.
    :param key: The key in the storage dict corresponding to a set of hooks.
    :param id: A unique identifier for the specific hook within the set.
    """

    def __init__(self, storage: nn.ModuleDict, key: str, id: str) -> None:
        self.storage_ref = weakref.ref(storage)
        self.key = key
        self.id = id

    def remove(self) -> None:
        """
        Removes the hook with the specified id from the storage, if it exists.
        If the set of hooks under a particular key becomes empty after removal, the key is also deleted.
        Steps:
        - Checks if the storage still exists (i.e., if it's not garbage collected).
        - Verifies if the key and id are present in the storage.
        - If hook exists, hook will removed.
        - If no hooks left under the key, the key is also removed from the storage.
        """
        storage: Optional[nn.ModuleDict] = self.storage_ref()
        if storage is not None and self.key in storage and self.id in storage[self.key]:
            del storage[self.key][self.id]
            if not storage[self.key]:
                del storage[self.key]


class HookStorage(nn.Module):
    """
    A module for storing and executing hooks.

    :param pre_hooks: A instance of nn.ModuleDict for storing pre-hooks.
    :param post_hooks: A instance of nn.ModuleDict for storing post-hooks.
    """

    def __init__(self) -> None:
        """
        Initialize an empty HookStorage.
        """
        super().__init__()
        self.pre_hooks: nn.ModuleDict = nn.ModuleDict()
        self.post_hooks: nn.ModuleDict = nn.ModuleDict()

    @staticmethod
    def _generate_key(op_name: str, port_id: int) -> str:
        """
        Return key for module dict that generated by operation name and port id.

        :param op_name: The operation name the hook is associated with.
        :param port_id: The port ID the hook is associated with.
        :return: Generate key for module dict.
        """
        return f"{op_name}__{port_id}"

    @staticmethod
    def _get_next_hook_id(hooks_dict: nn.ModuleDict) -> str:
        """
        Determines the next available hook ID by finding the highest existing hook ID and incrementing it.

        :param hooks_dict: A dictionary containing existing hooks.
        :return: The next available hook ID as a string. Starts from '0' if no hooks exist.
        """

        if not hooks_dict:
            return "0"
        return str(max([int(k) for k in hooks_dict]) + 1)

    @classmethod
    def _insert_hook(
        cls, storage_dict: nn.ModuleDict, op_name: str, port_id: int, hook: nn.Module
    ) -> RemovableHookHandle:
        """
        Inserts a hook into the storage under the appropriate key.

        :param storage_dict: The storage of hook.
        :param op_name: The operation name the hook is associated with.
        :param port_id: The port ID the hook is associated with.
        :param hook: The hook module to be stored.
        :return: A handle that can be used to remove the hook later.
        """

        hook_key = cls._generate_key(op_name, port_id)
        if hook_key not in storage_dict:
            storage_dict[hook_key] = nn.ModuleDict()

        hook_id = cls._get_next_hook_id(storage_dict[hook_key])
        storage_dict[hook_key][hook_id] = hook

        return RemovableHookHandle(storage_dict, hook_key, hook_id)

    def register_pre_function_hook(self, op_name: str, port_id: int, hook: nn.Module) -> RemovableHookHandle:
        """
        Registers a pre-function hook to be executed before a specific operation.

        :param op_name: The operation name the hook is associated with.
        :param port_id: The port ID the hook is associated with.
        :param hook: The pre-function hook to be stored.
        :return: A handle for removing the registered hook.
        """
        return self._insert_hook(self.pre_hooks, op_name, port_id, hook)

    def register_post_function_hook(self, op_name: str, port_id: int, hook: nn.Module) -> RemovableHookHandle:
        """
        Registers a post-function hook to be executed before a specific operation.

        :param op_name: The operation name the hook is associated with.
        :param port_id: The port ID the hook is associated with.
        :param hook: The pre-function hook to be stored.
        :return: A handle for removing the registered hook.
        """
        return self._insert_hook(self.post_hooks, op_name, port_id, hook)

    @classmethod
    def _execute_hooks(cls, storage_dict: nn.ModuleDict, op_name: str, port_id: int, value: Any) -> Any:
        """
        Executes all hooks of a given type (pre or post) for a specific operation and port,
        passing and potentially modifying a value.

        :param storage_dict: The storage of hook.
        :param op_name: The operation name the hooks are associated with.
        :param port_id: The port ID the hooks are associated with.
        :param value: The input value to be passed through the hooks.
        :return: The modified value after all hooks have been applied.
        """
        hook_key = cls._generate_key(op_name, port_id)
        if hook_key not in storage_dict:
            return value
        for hook in storage_dict[hook_key].values():
            value = hook(value)
        return value

    def execute_pre_function_hooks(self, op_name: str, port_id: int, value: Any) -> Any:
        """
        Executes all pre-function hooks for a given operation and port.

        :param op_name: The operation name the hooks are associated with.
        :param port_id: The port ID the hooks are associated with.
        :param value: The input value to be passed through the pre-function hooks.
        :return: The value after all pre-function hooks have been executed.
        """
        return self._execute_hooks(self.pre_hooks, op_name, port_id, value)

    def execute_post_function_hooks(self, op_name: str, port_id: int, value: Any) -> Any:
        """
        Executes all post-function hooks for a given operation and port.

        :param op_name: The operation name the hooks are associated with.
        :param port_id: The port ID the hooks are associated with.
        :param value: The input value to be passed through the pre-function hooks.
        :return: The value after all post-function hooks have been executed.
        """
        return self._execute_hooks(self.post_hooks, op_name, port_id, value)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.experimental.torch2.function_hook.graph.build_graph_mode import build_graph as build_graph
from nncf.experimental.torch2.function_hook.wrapper import get_hook_storage as get_hook_storage
from nncf.experimental.torch2.function_hook.wrapper import is_wrapped as is_wrapped
from nncf.experimental.torch2.function_hook.wrapper import register_post_function_hook as register_post_function_hook
from nncf.experimental.torch2.function_hook.wrapper import register_pre_function_hook as register_pre_function_hook
from nncf.experimental.torch2.function_hook.wrapper import wrap_model as wrap_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, TypeVar

from nncf import Dataset
from nncf.common.graph.graph import NNCFGraph
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.experimental.quantization.quantizer import Quantizer
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.quantization.algorithms.min_max.algorithm import MinMaxQuantization
from nncf.quantization.range_estimator import RangeEstimatorParameters

TModel = TypeVar("TModel")


class MinMaxRangeEstimator(Algorithm):
    def __init__(
        self,
        quantizer: Quantizer,
        subset_size: int = 300,
        inplace_statistics: bool = True,
        batchwise_statistics: bool = False,
        activations_range_estimator_params: Optional[RangeEstimatorParameters] = None,
        weights_range_estimator_params: Optional[RangeEstimatorParameters] = None,
    ):
        """
        :param quantizer: Instance of Quantizer to retrieve a quantization config
            for the given model.
        :param subset_size: Size of a subset to calculate activations statistics used
            for quantization, defaults to 300.
        :param inplace_statistics: Defines wheather to calculate quantizers statistics
            by backend graph operations or by default Python implementation, defaults
            to True.
        :param batchwise_statistics: Determines whether quantizer statistics should be calculated
            for each item of the batch or for the entire batch, default is False.
        :param activations_range_estimator_params: Quantization range estimation
            parameters for activation.
        :param weights_range_estimator_params: Quantization range estimation parameters
            for weights.
        """
        self._quantizer = quantizer
        self._min_max_algo = MinMaxQuantization(
            subset_size=subset_size,
            inplace_statistics=inplace_statistics,
            batchwise_statistics=batchwise_statistics,
            activations_range_estimator_params=activations_range_estimator_params,
            weights_range_estimator_params=weights_range_estimator_params,
        )

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.TORCH_FX]

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        if self._min_max_algo._quantization_target_points_to_qconfig is None:
            raise RuntimeError(
                "Statistic points are not available."
                " Please call `get_statistic_points` before calling the `apply` method."
            )
        return self._min_max_algo.apply(model=model, graph=graph, statistic_points=statistic_points)

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        quantizer_setup = self._quantizer.get_quantization_setup(model, graph)
        self._min_max_algo._set_backend_entity(model)
        self._min_max_algo._init_cache()
        self._min_max_algo.fill_quantization_target_points(quantizer_setup, graph)
        return self._min_max_algo.get_cached_statistic_points(model, graph)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
import warnings
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, cast

TObj = TypeVar("TObj")


def warning_deprecated(msg: str) -> None:
    """
    Display a warning message indicating that a certain functionality is deprecated.

    :param msg: The warning message to display.
    """
    # Note: must use FutureWarning in order not to get suppressed by default
    warnings.warn(msg, FutureWarning, stacklevel=2)


def deprecated(
    msg: Optional[str] = None, start_version: Optional[str] = None, end_version: Optional[str] = None
) -> Callable[[TObj], TObj]:
    """
    Decorator to mark a function or class as deprecated.

    :param msg: Message to provide additional information about the deprecation.
    :param start_version: Start version from which the function or class is deprecated.
    :param end_version: End version until which the function or class is deprecated.

    :return: The decorator function.
    """

    def decorator(obj: TObj) -> TObj:

        if isinstance(obj, types.FunctionType):

            @wraps(obj)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                name = f"function '{obj.__module__}.{obj.__name__}'"
                text = _generate_deprecation_message(name, msg, start_version, end_version)
                warning_deprecated(text)
                return obj(*args, **kwargs)

            return cast(TObj, wrapper)

        if isinstance(obj, type):
            original_init = obj.__init__  # type: ignore[misc]

            @wraps(original_init)
            def wrapped_init(*args: Any, **kwargs: Any) -> Any:
                name = f"class '{obj.__module__}.{obj.__name__}'"
                text = _generate_deprecation_message(name, msg, start_version, end_version)
                warning_deprecated(text)
                return original_init(*args, **kwargs)

            obj.__init__ = wrapped_init  # type: ignore[misc]

            return cast(TObj, obj)

        raise TypeError("The @deprecated decorator can only be used on functions or classes.")

    return decorator


def _generate_deprecation_message(
    name: str, text: Optional[str], start_version: Optional[str], end_version: Optional[str]
) -> str:
    """
    Generate a deprecation message for a given name, with optional start and end versions.

    :param name: The name of the deprecated feature.
    :param text: Additional text to include in the deprecation message.
    :param start_version: The version from which the feature is deprecated.
    :param end_version: The version in which the feature will be removed.
    :return: The deprecation message.
    """
    msg = (
        f"Usage of {name} is deprecated {f'starting from NNCF v{start_version} ' if start_version else ''}"
        f"and will be removed in {f'NNCF v{end_version}.' if end_version else 'a future NNCF version.'}"
    )
    if text:
        return "\n".join([msg, text])
    return msg



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple, Union

import torch

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import CONST_NOOP_METATYPES
from nncf.torch.dynamic_graph.context import PreHookId
from nncf.torch.external_hook import ExternalOpCallHook
from nncf.torch.graph import operator_metatypes as om
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.layers import AsymmetricQuantizer
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.layers import SymmetricQuantizer

CONV_META_TYPES = [
    om.PTConv1dMetatype,
    om.PTConv2dMetatype,
    om.PTConv3dMetatype,
    om.PTDepthwiseConv1dSubtype,
    om.PTDepthwiseConv2dSubtype,
    om.PTDepthwiseConv3dSubtype,
    om.PTConvTranspose1dMetatype,
    om.PTConvTranspose2dMetatype,
    om.PTConvTranspose3dMetatype,
]

OPERATORS_WITH_BIAS_METATYPES = CONV_META_TYPES
CONV_FUSED_META_TYPES = [om.PTBatchNormMetatype]


def find_const_node_in_constant_subgraph(node: NNCFNode, graph: NNCFGraph) -> Optional[NNCFNode]:
    """
    Finds a constant node within a constant subgraph, recursively traversing noop and quantize nodes.

    :param node: The starting node to search from.
    :param graph: The NNCFGraph.
    :return: The constant node found within the subgraph, or None if no constant node is found.
    """
    if node.metatype == om.PTNoopMetatype or node.node_type in om.QUANTIZE_NODE_TYPES:
        prev_nodes = graph.get_previous_nodes(node)
        if len(prev_nodes) != 1:
            return None
        return find_const_node_in_constant_subgraph(prev_nodes[0], graph)
    if node.metatype in CONST_NOOP_METATYPES:
        return node
    return None


def get_const_node(node: NNCFNode, port_id: int, graph: NNCFGraph) -> Optional[NNCFNode]:
    """
    Retrieves the constant node providing the input to a specific port of a given node in the NNCF graph.

    :param node: The NNCF node for which to find the constant input node.
    :param port_id: The ID of the input port to consider.
    :param graph: The NNCF graph containing the nodes.
    :return: The NNCF node providing the constant input to the specified port, or None if no such node is found.
    """
    for prev_node in graph.get_previous_nodes(node):
        edge = graph.get_edge(prev_node, node)
        if edge.input_port_id == port_id:
            weight_node = find_const_node_in_constant_subgraph(prev_node, graph)
            if weight_node is None:
                raise nncf.InternalError("Could not find a constant node in the model graph.")
            return weight_node


def split_const_name(const_name: str) -> Tuple[str, str]:
    """
    Splits the constant name into module and attribute names.

    :param const_name: The full name of the constant, including module and attribute.
    :return:
        - module_name: The name of the module containing the constant.
        - weight_attr_name: The name of the constant attribute within the module.
    """
    index = const_name.rfind(".")
    if index == -1:
        return str(), const_name
    module_name = const_name[:index]
    weight_attr_name = const_name[index + 1 :]
    return module_name, weight_attr_name


def get_module_by_name(module_name: str, model: torch.nn.Module) -> torch.nn.Module:
    """
    Retrieves a module from a PyTorch model by its hierarchical name.

    :param module_name: The name of the module to retrieve (e.g., "module1.submodule2").
    :param model: The model to search within.
    :return: The retrieved module.
    """
    if not module_name:
        return model
    curr_module = model
    for name in module_name.split("."):
        for child_name, child_module in curr_module.named_children():
            if child_name == name:
                curr_module = child_module
                break
        else:
            raise nncf.ModuleNotFoundError(f"Could not find the {module_name} module in the model.")
    return curr_module


def get_const_data(const_node: NNCFNode, model: NNCFNetwork) -> torch.Tensor:
    """
    Retrieves a constant tensor associated with a given node.

    :param const_node: The node associated with const data.
    :param model: The NNCFNetwork object.
    :return: A torch.Tensor object containing the constant value.
    """
    const_name = const_node.layer_attributes.name
    module_name, const_attr_name = split_const_name(const_name)
    module = get_module_by_name(module_name, model)
    data = getattr(module, const_attr_name)
    if isinstance(data, torch.nn.Parameter):
        return data.data
    return data


def get_const_data_on_port(node: NNCFNode, port_id: int, model: NNCFNetwork) -> torch.Tensor:
    """
    Retrieves a constant tensor associated with a given node and input port in an NNCF graph.

    :param node: The node to retrieve the constant from.
    :param port_id:  The port id within the node that holds the constant.
    :param model: The NNCFNetwork object.
    :return: A torch.Tensor object containing the constant value, or None if the constant is not found.
    """
    graph = model.nncf.get_graph()
    const_node = get_const_node(node, port_id, graph)
    if const_node is None:
        return None
    return get_const_data(const_node, model)


def get_potential_fused_node(node_name: str, nncf_graph: NNCFGraph) -> Optional[NNCFNode]:
    """
    Retrieves the next node in the NNCF graph that could be fused with the provided node during runtime optimization.

    :param node_name: The node name.
    :param nncf_graph: The NNCF graph.
    :return: The node that can be fused or None if no suitable node is found.
    """
    target_node = nncf_graph.get_node_by_name(node_name)

    if target_node.metatype in CONV_META_TYPES:
        next_nodes = nncf_graph.get_next_nodes(target_node)
        for node in next_nodes:
            if node.metatype in CONV_FUSED_META_TYPES:
                return node
    return None


def is_node_with_fused_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
    """
    Checks if the node has a fused bias.

    :param node: The node to check.
    :param nncf_graph: The NNCF graph.
    :return: Return `True` if `node` corresponds to the operation
        with bias (bias is added to the output tensor of that operation),
        `False` otherwise.
    """
    if node.metatype not in OPERATORS_WITH_BIAS_METATYPES:
        return False
    fused_node = get_potential_fused_node(node.node_name, nncf_graph)
    if fused_node is not None:
        node = fused_node
    bias_port = node.metatype.bias_port_id
    bias = get_const_node(node, bias_port, nncf_graph)
    return bias is not None


def get_fused_bias_value(node: NNCFNode, model: NNCFNetwork) -> Optional[torch.Tensor]:
    """
    Returns the bias tensor for the node or for potential fused node.

    :param node: The node that corresponds to the operation with bias.
    :param model: The model that contains this operation.
    :return: The bias value that is applied to the output tensor of the node's operation.
    """
    nncf_graph = model.nncf.get_graph()
    fused_node = get_potential_fused_node(node.node_name, nncf_graph)
    bias = get_const_data_on_port(node, node.metatype.bias_port_id, model)

    if fused_node is None:
        return bias

    fused_bias = get_const_data_on_port(fused_node, fused_node.metatype.bias_port_id, model)
    if bias is None:
        return fused_bias

    fused_weight = get_const_data_on_port(fused_node, fused_node.metatype.weight_port_ids[0], model)
    return bias * fused_weight + fused_bias


def update_fused_bias(target_node_name: str, new_bias: torch.Tensor, model: NNCFNetwork) -> None:
    """
    Update bias for target module or potential fused module.

    :param target_node_name: The target node name.
    :param new_bias: New bias value.
    :param model: The model.
    """
    nncf_graph = model.nncf.get_graph()
    target_node = nncf_graph.get_node_by_name(target_node_name)
    fused_node = get_potential_fused_node(target_node_name, nncf_graph)
    if fused_node is None:
        set_const_data_to_port_id(new_bias, target_node, target_node.metatype.bias_port_id, model)
        return

    target_bias_node = get_const_node(target_node, target_node.metatype.bias_port_id, nncf_graph)
    fused_bias_node = get_const_node(fused_node, fused_node.metatype.bias_port_id, nncf_graph)
    fused_weight_node = get_const_node(fused_node, fused_node.metatype.weight_port_ids[0], nncf_graph)

    if target_bias_node is None:
        set_const_data(new_bias, fused_bias_node, model)
        return

    new_bias = new_bias - get_const_data(target_bias_node, model) * get_const_data(fused_weight_node, model)
    set_const_data(new_bias, fused_bias_node, model)


def get_weight_tensor_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[int]:
    """
    Returns list of input port ids that contains traced constant tensor.

    :param node: Target node that contains weights.
    :param graph: The NNCF graph.
    :return: List of ports with weights.
    """
    weight_port_ids = []
    for edge in graph.get_input_edges(node):
        if edge.input_port_id in node.metatype.weight_port_ids:
            weight_node = find_const_node_in_constant_subgraph(edge.from_node, graph)
            if weight_node:
                weight_port_ids.append(edge.input_port_id)
    return weight_port_ids


def set_const_data(data: torch.Tensor, const_node: NNCFNode, model: NNCFNetwork) -> None:
    """
    Sets the constant data associated with a specific constant node in an NNCF network model.

    :param data: The constant data tensor to be set.
    :param const_node: The NNCF node representing the constant data.
    :param model: The NNCF network model.
    """
    const_name = const_node.layer_attributes.name
    module_name, const_attr_name = split_const_name(const_name)
    module = get_module_by_name(module_name, model)
    const = getattr(module, const_attr_name)
    if isinstance(const, torch.nn.Parameter):
        const.data = data
    else:
        setattr(module, const_attr_name, data)


def set_const_data_to_port_id(data: torch.Tensor, node: NNCFNode, port_id: int, model: NNCFNetwork) -> None:
    """
    Sets the value of a constant tensor within a specified node in an NNCFNetwork.

    :param data: The tensor containing the new value to be set for the constant.
    :param node: The NNCF node representing the operation that uses the constant.
    :param const_port_id: The input port id of the node that receives the constant.
    :param model: The NNCF network containing the module to be modified.
    """
    graph = model.nncf.get_graph()
    const_node = get_const_node(node, port_id, graph)
    if const_node is None:
        raise nncf.InternalError(f"No found node with constant for {node.node_name} on {port_id} port")
    const_name = const_node.layer_attributes.name
    module_name, const_attr_name = split_const_name(const_name)
    module = get_module_by_name(module_name, model)
    const = getattr(module, const_attr_name)
    if isinstance(const, torch.nn.Parameter):
        const.data = data
    else:
        setattr(module, const_attr_name, data)


def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
    """
    Check that module have fake_quantizer for its weights (supports only metatypes with only one weight port).

    :param node: The target node.
    :param nncf_graph: The NNCF graph.
    :return bool: return `True` if the node is quantized.
    """
    assert len(node.metatype.weight_port_ids) == 1, "Support only metatype with only 1 weighted port"
    for edge in nncf_graph.get_input_edges(node):
        if edge.input_port_id in node.metatype.weight_port_ids and edge.from_node.node_type in om.QUANTIZE_NODE_TYPES:
            return True
    return False


def get_fake_quantizer(
    node: NNCFNode, port_id: Optional[int], model: NNCFNetwork
) -> Union[SymmetricQuantizer, AsymmetricQuantizer]:
    """
    Retrieves the fake quantizer associated with a specific node and input port id.

    :param node: The NNCFNode representing the node for which to retrieve the quantizer.
    :param port_id: The port id number for which to retrieve the quantizer module, None means output port.
    :param model: The NNCFNetwork instance.
    :return: Fake Quantizer module if exists, overwise None.
    """

    address_map = model.nncf.get_node_to_op_address_mapping()
    op_addr = address_map[node.node_name]

    if port_id is not None:
        id = PreHookId(op_address=op_addr, input_port_id=port_id)
        hook_container = model.nncf._compressed_context._pre_hooks.get(id, {})
    else:
        hook_container = model.nncf._compressed_context._post_hooks.get(op_addr, {})

    for call_hook in hook_container.values():
        if isinstance(call_hook, ExternalOpCallHook):
            storage = getattr(model.nncf, call_hook._storage_name)
            module = storage[call_hook._storage_key]
            if isinstance(module, BaseQuantizer):
                return module
    return None


def get_weight_channel_axes(metatype: om.PTOperatorMetatype, ndims: int, input_port_id: int) -> Tuple[int, ...]:
    """
    Returns axes numbers of the weight tensor which correspond to its channels.

    :param metatype: The node metatype for which the target dimension is being determined.
    :param input_port_id: The input port id.
    :return: The target dimension for weight compression.
    """
    if metatype == om.PTAddmmMetatype:
        if input_port_id == 1:
            return (ndims - 2,)
        if input_port_id == 2:
            return (ndims - 1,)
        raise ValueError(f"Unexpected {input_port_id=} for {metatype=}")
    if metatype == om.PTMatMulMetatype:
        if input_port_id == 0:
            return () if ndims < 2 else (ndims - 2,)
        if input_port_id == 1:
            return () if ndims < 2 else (ndims - 1,)
        raise ValueError(f"Unexpected {input_port_id=} for {metatype=}")
    if metatype in [om.PTConvTranspose1dMetatype, om.PTConvTranspose2dMetatype, om.PTConvTranspose3dMetatype]:
        return (1,)
    return (0,)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Since we are not reading XML, but creating it, the package security message is irrelevant
import xml.etree.ElementTree as ET  # nosec
from typing import Callable, Dict, List, Optional, Tuple

from nncf.common.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.layer_attributes import Dtype


class Tags:
    NET = "net"
    NODES = "layers"
    EDGES = "edges"
    NODE = "layer"
    EDGE = "edge"
    DATA = "data"
    INPUT = "input"
    OUTPUT = "output"
    PORT = "port"
    DIM = "dim"


class PortDesc:
    """
    Represents a port description in the computational graph.
    """

    def __init__(self, port_id: str, precision: str, shape: Optional[List[int]] = None):
        """
        :param port_id: The identifier of the port.
        :param precision: Precision of the tensor corresponding to the port, either "fp32" or "i32".
        :param shape: The shape of the tensor. Defaults to an empty list if not provided.
        """
        self.port_id = port_id
        if shape is None:
            shape = []
        self.shape = shape
        self.precision = precision

    def as_xml_element(self) -> ET.Element:
        """
        Converts the PortDesc object into an XML element.

        :return: The Element representing the port in XML
        """
        port = ET.Element(Tags.PORT, id=self.port_id, precision=self.precision)

        for i in self.shape:
            dim = ET.Element(Tags.DIM)
            dim.text = str(i)
            port.append(dim)
        return port


class NodeDesc:
    """
    Represents a node description in the computational graph.
    """

    def __init__(
        self,
        node_id: str,
        name: str,
        node_type: str,
        attrs: Optional[Dict[str, str]] = None,
        inputs: Optional[List[PortDesc]] = None,
        outputs: Optional[List[PortDesc]] = None,
    ):
        """
        :param node_id: The identifier of the node.
        :param name: The name of the node.
        :param type: The type of the node.
        :param attrs: Additional attributes of the node. Default empty dictionary.
        :param inputs: List of input ports for the node.
        :param outputs: List of output ports for the node.
        """
        self.node_id = node_id
        self.name = name
        self.type = node_type
        if attrs is None:
            attrs = {}
        self.attrs = attrs
        self.inputs = inputs
        self.outputs = outputs

    def as_xml_element(self) -> ET.Element:
        """
        Converts the NodeDesc object into an XML element.

        :return: The Element representing the node in XML
        """
        node = ET.Element(Tags.NODE, id=self.node_id, name=self.name, type=self.type)
        ET.SubElement(node, Tags.DATA, self.attrs)

        if self.inputs:
            input_ = ET.SubElement(node, Tags.INPUT)
            for port in self.inputs:
                input_.append(port.as_xml_element())

        if self.outputs:
            output = ET.SubElement(node, Tags.OUTPUT)
            for port in self.outputs:
                output.append(port.as_xml_element())

        return node


class EdgeDesc:
    """
    Represents an edge description in the computational graph.
    """

    def __init__(self, from_node: str, from_port: str, to_node: str, to_port: str):
        """
        :param from_node: The identifier of the source node.
        :param from_port: The identifier of the output port of the source node.
        :param to_node: The identifier of the target node.
        :param to_port: The identifier of the input port of the target node.
        """
        self.from_node = from_node
        self.from_port = from_port
        self.to_node = to_node
        self.to_port = to_port

    def as_xml_element(self) -> ET.Element:
        """
        Converts the EdgeDesc object into an XML element.

        :return: The Element representing the edge in XML
        """
        attrs = {
            "from-layer": self.from_node,
            "from-port": self.from_port,
            "to-layer": self.to_node,
            "to-port": self.to_port,
        }
        edge = ET.Element(Tags.EDGE, attrs)
        return edge


GET_ATTRIBUTES_FN_TYPE = Callable[[NNCFNode], Dict[str, str]]


def convert_nncf_dtype_to_ov_dtype(dtype: Dtype) -> str:
    """
    Converts a nncf dtype to an openvino dtype string.

    :param dtype: The data type to be converted.
    :return: The openvino dtype string corresponding to the given data type.
    """

    dummy_precision_map: Dict[Dtype, str] = {Dtype.INTEGER: "i32", Dtype.FLOAT: "f32"}

    return dummy_precision_map[dtype]


def get_graph_desc(
    graph: NNCFGraph, include_fq_params: bool = False, get_attributes_fn: Optional[GET_ATTRIBUTES_FN_TYPE] = None
) -> Tuple[List[NodeDesc], List[EdgeDesc]]:
    """
    Retrieves descriptions of nodes and edges from an NNCFGraph.

    :param graph: The NNCFGraph instance to extract descriptions from.
    :param include_fq_params: Whether to include FakeQuantize parameters in the description.
    :param get_attributes_fn: A function to retrieve additional attributes for nodes.
        Defaults to a function returning {"metatype": str(x.metatype.name)}.
    :return: A tuple containing lists of NodeDesc and EdgeDesc objects
        representing the nodes and edges of the NNCFGraph.
    """

    if get_attributes_fn is None:
        get_attributes_fn = lambda x: {
            "metatype": str(x.metatype.name),
        }
    include_node: Dict[int, bool] = {}
    edges = []
    for edge in graph.get_all_edges():
        if not include_fq_params and edge.to_node.node_type == "FakeQuantize" and edge.input_port_id != 0:
            include_node[edge.from_node.node_id] = False
            continue

        edges.append(
            EdgeDesc(
                from_node=str(edge.from_node.node_id),
                from_port=str(edge.output_port_id),
                to_node=str(edge.to_node.node_id),
                to_port=str(edge.input_port_id),
            )
        )

    nodes = []
    for node in graph.get_all_nodes():
        if not include_node.get(node.node_id, True):
            continue

        inputs = []
        for edge in graph.get_input_edges(node):
            if not include_fq_params and node.node_type == "FakeQuantize" and edge.input_port_id != 0:
                continue
            inputs.append(
                PortDesc(
                    port_id=str(edge.input_port_id),
                    precision=convert_nncf_dtype_to_ov_dtype(edge.dtype),
                    shape=edge.tensor_shape,
                )
            )

        outputs = []
        for edge in graph.get_output_edges(node):
            outputs.append(
                PortDesc(
                    port_id=str(edge.output_port_id),
                    precision=convert_nncf_dtype_to_ov_dtype(edge.dtype),
                    shape=edge.tensor_shape,
                )
            )

        nodes.append(
            NodeDesc(
                node_id=str(node.node_id),
                name=node.node_name,
                node_type=node.node_type.title(),
                attrs=get_attributes_fn(node),
                inputs=inputs,
                outputs=outputs,
            )
        )

    return nodes, edges


def save_for_netron(
    graph: NNCFGraph,
    save_path: str,
    graph_name: str = "Graph",
    include_fq_params: bool = False,
    get_attributes_fn: Optional[GET_ATTRIBUTES_FN_TYPE] = None,
):
    """
    Save the NNCFGraph information in an XML file suitable for visualization with Netron.

    :param graph: The NNCFGraph instance to visualize.
    :param save_path: The path to save the Netron-compatible file.
    :param graph_name: The name of the graph. Defaults to "Graph".
    :param include_fq_params: Whether to include FakeQuantize parameters in the visualization.
    :param get_attributes_fn: A function to retrieve additional attributes for nodes.
        Defaults to a function returning {"metatype": str(x.metatype.name)}.
    """

    node_descs, edge_descs = get_graph_desc(graph, include_fq_params, get_attributes_fn)

    net = ET.Element(Tags.NET, name=graph_name)

    nodes = ET.SubElement(net, Tags.NODES)
    for node in node_descs:
        nodes.append(node.as_xml_element())

    edges = ET.SubElement(net, Tags.EDGES)
    for edge in edge_descs:
        edges.append(edge.as_xml_element())

    # ET.indent(net)  # Only Python 3.9
    with open(save_path, "wb") as f:
        f.write(ET.tostring(net))



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

from nncf.api.compression import CompressionAlgorithmController


class SparsityController(CompressionAlgorithmController):
    """
    This is the class from which all sparsity controllers inherit.
    """

    def set_sparsity_level(self, sparsity_level: float) -> None:
        """
        Sets the sparsity level that should be applied to the model's weights.

        :param sparsity_level: Sparsity level that should be applied to the model's weights.
        """

    def freeze(self) -> None:
        """
        Freezes all sparsity masks. Sparsity masks will not be trained after calling this method.
        """

    @property
    @abstractmethod
    def current_sparsity_level(self) -> float:
        """
        Returns the current sparsity level of the underlying model.
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch

from nncf.torch.dynamic_graph.patch_pytorch import register_operator
from nncf.torch.functions import STThreshold


@register_operator()
def binary_mask_by_threshold(
    input_tensor: torch.Tensor, threshold: float = 0.5, max_percentile: float = 0.98
) -> torch.Tensor:
    """
    Conduct straight-through thresholding function while limiting the maximum threshold.

    :param input_tensor: The tensor to conduct thresholding.
    :param threshold: The criterion for thresholding.
    :param max_percentile: Specifies the `q`-th quantiles of the input tensor.
    :return: The mask for the input tensor with the threshold limited by `max_percentile`.
    """
    with torch.no_grad():
        max_threshold = torch.quantile(input_tensor, q=max_percentile).item()
    return STThreshold.apply(input_tensor, min(threshold, max_threshold))
